#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import six

import yaql
from yaql.language import exceptions
from yaql.language import utils


class Expression(object):
    def __call__(self, receiver, context, engine):
        pass


@six.python_2_unicode_compatible
class Function(Expression):
    def __init__(self, name, *args):
        self.name = name
        self.args = args
        self.uses_receiver = True

    def __call__(self, receiver, context, engine):
        return context(self.name, engine, receiver, context)(*self.args)

    def __str__(self):
        return u'{0}({1})'.format(self.name, ', '.join(
            map(six.text_type, self.args)))


class BinaryOperator(Function):
    def __init__(self, op, obj1, obj2, alias):
        if alias is None:
            func_name = '#operator_' + op
        else:
            func_name = '*' + alias
        self.operator = op
        super(BinaryOperator, self).__init__(func_name, obj1, obj2)
        self.uses_receiver = False


class UnaryOperator(Function):
    def __init__(self, op, obj, alias):
        if alias is None:
            func_name = '#unary_operator_' + op
        else:
            func_name = '*' + alias
        self.operator = op
        super(UnaryOperator, self).__init__(func_name, obj)
        self.uses_receiver = False


class IndexExpression(Function):
    def __init__(self, value, *args):
        super(IndexExpression, self).__init__('#indexer', value, *args)
        self.uses_receiver = False


class ListExpression(Function):
    def __init__(self, *args):
        super(ListExpression, self).__init__('#list', *args)
        self.uses_receiver = False


class MapExpression(Function):
    def __init__(self, *args):
        super(MapExpression, self).__init__('#map', *args)
        self.uses_receiver = False


@six.python_2_unicode_compatible
class GetContextValue(Function):
    def __init__(self, path):
        super(GetContextValue, self).__init__('#get_context_data', path)
        self.path = path
        self.uses_receiver = False

    def __str__(self):
        return self.path.value


@six.python_2_unicode_compatible
class Constant(Expression):
    def __init__(self, value):
        self.value = value
        self.uses_receiver = False

    def __str__(self):
        if isinstance(self.value, six.text_type):
            return u"'{0}'".format(self.value)
        return six.text_type(self.value)

    def __call__(self, receiver, context, engine):
        return self.value


class KeywordConstant(Constant):
    pass


@six.python_2_unicode_compatible
class Wrap(Expression):
    def __init__(self, expression):
        self.expr = expression
        self.uses_receiver = False

    def __str__(self):
        return str(self.expr)

    def __call__(self, receiver, context, engine):
        return self.expr(receiver, context, engine)


@six.python_2_unicode_compatible
class MappingRuleExpression(Expression):
    def __init__(self, source, destination):
        self.source = source
        self.destination = destination
        self.uses_receiver = False

    def __str__(self):
        return u'{0} => {1}'.format(self.source, self.destination)

    def __call__(self, receiver, context, engine):
        return utils.MappingRule(
            self.source(receiver, context, engine),
            self.destination(receiver, context, engine))


@six.python_2_unicode_compatible
class Statement(Function):
    def __init__(self, expression, engine):
        self.expression = expression
        self.uses_receiver = False
        self.engine = engine
        super(Statement, self).__init__('#finalize', expression)

    def __call__(self, receiver, context, engine):
        if not context.collect_functions('#finalize'):
            context = context.create_child_context()
            context.register_function(lambda x: x, name='#finalize')
        try:
            return super(Statement, self).__call__(receiver, context, engine)
        except exceptions.WrappedException as e:
            six.reraise(type(e.wrapped), e.wrapped, sys.exc_info()[2])

    def evaluate(self, data=utils.NO_VALUE, context=None):
        if context is None or context is utils.NO_VALUE:
            context = yaql.create_context()
        if data is not utils.NO_VALUE:
            if self.engine.options.get('yaql.convertInputData', True):
                context['$'] = utils.convert_input_data(data)
            else:
                context['$'] = data
        return self(utils.NO_VALUE, context, self.engine)

    def __str__(self):
        return str(self.expression)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
Common module describes comparison operators for different types. Comparing
with null value is considered separately.
"""

from yaql.language import specs


@specs.name('*equal')
def eq(left, right):
    """:yaql:operator =

    Returns true if left and right are equal, false otherwise.

    It is system function and can be used to override behavior
    of comparison between objects.
    """
    return left == right


@specs.name('*not_equal')
def neq(left, right):
    """:yaql:operator !=

    Returns true if left and right are not equal, false otherwise.

    It is system function and can be used to override behavior
    of comparison between objects.
    """
    return left != right


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_<')
def left_lt_null(left, right):
    """:yaql:operator <

    Returns false. This function is called when left is not null and
    right is null.

    :signature: left < right
    :arg left: left operand
    :argType left: not null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> 1 < null
        false
    """
    return False


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_<=')
def left_lte_null(left, right):
    """:yaql:operator <=

    Returns false. This function is called when left is not null
    and right is null.

    :signature: left <= right
    :arg left: left operand
    :argType left: not null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> 1 <= null
        false
    """
    return False


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_>')
def left_gt_null(left, right):
    """:yaql:operator >

    Returns true. This function is called when left is not null
    and right is null.

    :signature: left > right
    :arg left: left operand
    :argType left: not null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> 1 > null
        true
    """
    return True


@specs.parameter('right', type(None), nullable=True)
@specs.parameter('left', nullable=False)
@specs.name('#operator_>=')
def left_gte_null(left, right):
    """:yaql:operator >=

    Returns true. This function is called when left is not null
    and right is null.

    :signature: left >= right
    :arg left: left operand
    :argType left: not null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> 1 >= null
        true
    """
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_<')
def null_lt_right(left, right):
    """:yaql:operator <

    Returns true. This function is called when left is null and
    right is not.

    :signature: left < right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: not null
    :returnType: boolean

    .. code:

        yaql> null < 2
        true
    """
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_<=')
def null_lte_right(left, right):
    """:yaql:operator <=

    Returns true. This function is called when left is null and
    right is not.

    :signature: left <= right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: not null
    :returnType: boolean

    .. code:

        yaql> null <= 2
        true
    """
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_>')
def null_gt_right(left, right):
    """:yaql:operator >

    Returns false. This function is called when left is null and right
    is not.

    :signature: left > right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: not null
    :returnType: boolean

    .. code:

        yaql> null > 2
        false
    """
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', nullable=False)
@specs.name('#operator_>=')
def null_gte_right(left, right):
    """:yaql:operator >=

    Returns false. This function is called when left is null and
    right is not.

    :signature: left >= right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: not null
    :returnType: boolean

    .. code:

        yaql> null >= 2
        false
    """
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_<')
def null_lt_null(left, right):
    """:yaql:operator <

    Returns false. This function is called when left and right are null.

    :signature: left < right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> null < null
        false
    """
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_<=')
def null_lte_null(left, right):
    """:yaql:operator <=

    Returns true. This function is called when left and right are null.

    :signature: left <= right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> null <= null
        true
    """
    return True


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_>')
def null_gt_null(left, right):
    """:yaql:operator >

    Returns false. This function is called when left and right are null.

    :signature: left > right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> null > null
        false
    """
    return False


@specs.parameter('left', type(None), nullable=True)
@specs.parameter('right', type(None), nullable=True)
@specs.name('#operator_>=')
def null_gte_null(left, right):
    """:yaql:operator >=

    Returns true. This function is called when left and right are null.

    :signature: left >= right
    :arg left: left operand
    :argType left: null
    :arg right: right operand
    :argType right: null
    :returnType: boolean

    .. code:

        yaql> null >= null
        true
    """
    return True


def register(context):
    context.register_function(eq)
    context.register_function(neq)
    context.register_function(left_lt_null)
    context.register_function(left_lte_null)
    context.register_function(left_gt_null)
    context.register_function(left_gte_null)

    context.register_function(null_lt_right)
    context.register_function(null_lte_right)
    context.register_function(null_gt_right)
    context.register_function(null_gte_right)

    context.register_function(null_lt_null)
    context.register_function(null_lte_null)
    context.register_function(null_gt_null)
    context.register_function(null_gte_null)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import exceptions
import yaql.tests


class TestCollections(yaql.tests.TestCase):
    def test_list(self):
        self.assertEqual([], self.eval('list()'))
        self.assertEqual([1, 2, 3], self.eval('list(1, 2, 3)'))
        self.assertEqual([1, 2, [3, 4]], self.eval('list(1, 2, list(3, 4))'))

    def test_list_expr(self):
        self.assertEqual([1, 2, 3], self.eval('[1,2,3]'))
        self.assertEqual([2, 4], self.eval('[1,[2]][1] + [3, [4]][1]'))
        self.assertEqual([1, 2, 3, 4], self.eval('[1,2] + [3, 4]'))
        self.assertEqual(2, self.eval('([1,2] + [3, 4])[1]'))
        self.assertEqual([], self.eval('[]'))

    def test_list_from_iterator(self):
        iterator = (i for i in range(3))
        self.assertEqual([0, 1, 2], self.eval('list($)', data=iterator))

    def test_to_list(self):
        data = (i for i in range(3))
        self.assertEqual([0, 1, 2], self.eval('$.toList()', data=data))

        data = [0, 1, 2]
        self.assertEqual([0, 1, 2], self.eval('$.toList()', data=data))

        data = (0, 1, 2)
        self.assertEqual([0, 1, 2], self.eval('$.toList()', data=data))

        data = {'a': 1, 'b': 2}

        self.assertTrue(self.eval('isList($.keys().toList())', data=data))

    def test_list_concatenates_and_flatten_generators(self):
        generator_func = lambda: (i for _ in range(2) for i in range(3))

        self.context['$seq1'] = generator_func()
        self.context['$seq2'] = generator_func()

        self.assertEqual([0, 1, 2, 0, 1, 2, 0, 1, 2, 0, 1, 2],
                         self.eval('list($seq1, $seq2)'))

    def test_indexer_list_access(self):
        data = [1, 2, 3]
        self.assertEqual(1, self.eval('$[0]', data=data))
        self.assertEqual(3, self.eval('$[-1]', data=data))
        self.assertEqual(2, self.eval('$[-1-1]', data=data))
        self.assertRaises(IndexError,
                          self.eval, "$[3]", data=data)
        self.assertRaises(IndexError,
                          self.eval, "$[-4]", data=data)
        self.assertRaises(exceptions.NoMatchingFunctionException,
                          self.eval, "$[a]", data=data)

    def test_dict(self):
        self.assertEqual(
            {'b c': 13, 'a': 2, 4: 5, None: None, True: False, 8: 8},
            self.eval("dict(a => 2, 'b c' => 13, 4 => 5, "
                      "null => null, true => false, 2+6=>8)"))

    def test_dict_expr(self):
        self.assertEqual(
            {'b c': 13, 'a': 2, 4: 5, None: None, True: False, 8: 8},
            self.eval("{a => 2, 'b c' => 13, 4 => 5, "
                      "null => null, true => false, 2+6=>8}"))

        self.assertEqual({'b': 2, 'a': 1}, self.eval('{a => 1} + {b=>2}'))
        self.assertEqual({}, self.eval('{}'))

    def test_dict_from_sequence(self):
        self.assertEqual({'a': 1, 'b': 2},
                         self.eval("dict(list(list(a, 1), list('b', 2)))"))

        generator = (i for i in (('a', 1), ['b', 2]))
        self.assertEqual({'a': 1, 'b': 2},
                         self.eval('dict($)', data=generator))

    def test_to_dict(self):
        self.assertEqual({1: 1, 2: 4, 3: 9},
                         self.eval('$.toDict($, $*$)', data=[1, 2, 3]))

    def test_keyword_dict_access(self):
        data = {'A': 12, 'b c': 44, '__d': 99, '_e': 999}
        self.assertEqual(12, self.eval('$.A', data=data))
        self.assertEqual(999, self.eval('$._e', data=data))

        self.assertRaises(exceptions.NoMatchingFunctionException,
                          self.eval, "$.'b c'", data=data)
        self.assertRaises(exceptions.NoMatchingFunctionException,
                          self.eval, '$.123', data=data)
        self.assertRaises(KeyError, self.eval, '$.b', data=data)
        self.assertRaises(
            exceptions.YaqlLexicalException,
            self.eval, '$.__d', data=data)

    def test_indexer_dict_access(self):
        data = {'a': 12, 'b c': 44}
        self.assertEqual(12, self.eval('$[a]', data=data))
        self.assertEqual(44, self.eval("$['b c']", data=data))
        self.assertRaises(KeyError,
                          self.eval, "$[b]", data=data)

    def test_indexer_dict_access_with(self):
        data = {'a': 12, 'b c': 44}
        self.assertEqual(55, self.eval('$[c, 55]', data=data))
        self.assertEqual(66, self.eval('$[c, default => 66]', data=data))

    def test_list_eq(self):
        self.assertTrue(self.eval('[c, 55]=[c, 55]'))
        self.assertFalse(self.eval('[c, 55]=[55, c]'))
        self.assertFalse(self.eval('[c, 55]=null'))
        self.assertFalse(self.eval('null = [c, 55]'))

    def test_list_neq(self):
        self.assertFalse(self.eval('[c, 55] != [c, 55]'))
        self.assertTrue(self.eval('[c, 55] != [55, c]'))
        self.assertTrue(self.eval('[c, 55] != null'))
        self.assertTrue(self.eval('null != [c, 55]'))

    def test_dict_eq(self):
        self.assertTrue(self.eval('{a => [c, 55]} = {a => [c, 55]}'))
        self.assertTrue(self.eval('{[c, 55] => a} = {[c, 55] => a}'))
        self.assertFalse(self.eval('{[c, 55] => a} = {[c, 56] => a}'))
        self.assertFalse(self.eval('{[c, 55] => a, b => 1} = {[c, 55] => a}'))
        self.assertFalse(self.eval('{[c, 55] => a} = null'))

    def test_dict_neq(self):
        self.assertFalse(self.eval('{a => [c, 55]} != {a => [c, 55]}'))
        self.assertFalse(self.eval('{[c, 55] => a} != {[c, 55] => a}'))
        self.assertTrue(self.eval('{[c, 55] => a} != {[c, 56] => a}'))
        self.assertTrue(self.eval('{[c, 55] => a, b => 1} != {[c, 55] => a}'))
        self.assertTrue(self.eval('{[c, 55] => a} != null'))

    def test_dict_get(self):
        data = {'a': 12, 'b c': 44}
        self.assertEqual(12, self.eval('$.get(a)', data=data))
        self.assertIsNone(self.eval('$.get(b)', data=data))
        self.assertEqual(50, self.eval('$.get(c, 50)', data=data))

    def test_dict_set(self):
        data = {'a': 12, 'b c': 44}
        self.assertEqual(
            {'a': 99, 'b c': 44, 'x': None},
            self.eval('$.set(a, 99).set(x, null)', data=data))
        self.assertEqual(data, {'a': 12, 'b c': 44})

    def test_dict_set_many(self):
        data = {'a': 12, 'b c': 44}
        self.assertEqual(
            {'a': 55, 'b c': 44, 'd x': 99, None: None},
            self.eval('$.set(dict(a => 55, "d x" => 99, null => null))',
                      data=data))
        self.assertEqual(data, {'a': 12, 'b c': 44})

    def test_dict_set_many_inline(self):
        data = {'a': 12, 'b c': 44}
        self.assertEqual(
            {'a': 55, 'b c': 44, 'd x': 99},
            self.eval('$.set(a => 55, "d x" => 99)', data=data))
        self.assertEqual(data, {'a': 12, 'b c': 44})

    def test_dict_keys(self):
        data = {'a': 12, 'b': 44}
        self.assertItemsEqual(['a', 'b'], self.eval('$.keys()', data=data))

    def test_dict_values(self):
        data = {'a': 12, 'b': 44}
        self.assertItemsEqual([12, 44], self.eval('$.values()', data=data))

    def test_dict_items(self):
        data = {'a': 12, 'b': 44}
        self.assertItemsEqual([['a', 12], ['b', 44]],
                              self.eval('$.items()', data=data))
        self.assertEqual(data, self.eval('dict($.items())', data=data))

    def test_in(self):
        data = {'a': 12, 'b': 44}
        self.assertTrue(self.eval('44 in $.values()', data=data))
        self.assertTrue(self.eval('24 in $.values().select(2*$)', data=data))
        self.assertTrue(self.eval('{a => 2} in {{a => 2} => {b => 3}}.keys()'))
        self.assertTrue(self.eval('5 in set(1, 2, 5)'))
        self.assertTrue(self.eval('[1, 2] in set([1, 2], 5)'))
        self.assertTrue(self.eval('5 in [1, 2, 5]'))
        self.assertTrue(self.eval('[1, 2] in {[1, 2] => [3, 4]}.keys()'))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'a in $', data=data)

    def test_contains(self):
        data = {'a': 12, 'b': 44}
        self.assertTrue(self.eval('$.containsKey(a)', data=data))
        self.assertTrue(self.eval('$.values().contains(44)', data=data))
        self.assertFalse(self.eval('$.containsKey(null)', data=data))
        self.assertTrue(self.eval(
            '$.values().select(2*$).contains(24)', data=data))
        self.assertTrue(self.eval('set(1, 2, 5).contains(5)'))
        self.assertTrue(self.eval('[1, 2, 5].contains(5)'))
        self.assertTrue(
            self.eval('{{a => 2} => {b => 3}}.containsKey({a => 2})'))
        self.assertTrue(self.eval('{[1, 2] => [3, 4]}.containsKey([1, 2])'))
        self.assertTrue(self.eval('{[1, 2] => [3, 4]}.containsValue([3, 4])'))
        self.assertTrue(self.eval('set([1, 2], 5).contains([1, 2])'))

    def test_list_addition(self):
        self.assertEqual(
            [1, 2, 3, 4],
            self.eval('list(1, 2) + list(3, 4)'))
        self.assertEqual(
            [1, 2, 6, 8],
            self.eval('list(1, 2) + list(3, 4).select($ * 2)'))

    def test_dict_addition(self):
        self.assertEqual(
            {'a': 1, 'b': 2},
            self.eval('dict(a => 1) + dict(b => 2)'))

    def test_list_multiplication(self):
        self.assertItemsEqual(
            [1, 2, 1, 2, 1, 2],
            self.eval('3 * [1, 2]'))

        self.assertItemsEqual(
            [1, 2, 1, 2, 1, 2],
            self.eval('[1, 2] * 3'))

    def test_dict_list_key(self):
        self.assertEqual(
            3,
            self.eval('dict($ => 3).get($)', data=[1, 2]))

        self.assertEqual(
            3,
            self.eval('dict($ => 3).get($)', data=[1, [2]]))

    def test_dict_dict_key(self):
        self.assertEqual(
            3,
            self.eval('dict($ => 3).get($)', data={'a': 1}))

    def test_delete(self):
        self.assertEqual(
            [2, 3, 4],
            self.eval('[1, 2, 3, 4].delete(0)'))

        self.assertEqual(
            [3, 4],
            self.eval('[1, 2, 3, 4].delete(0, 2)'))

        self.assertEqual(
            [4],
            self.eval('[1, 2, 3, 4].delete(0, 2).delete(0)'))

        self.assertEqual(
            [1],
            self.eval('[1, 2, 3, 4].delete(1, -1)'))

        self.assertEqual(
            [1, 2, 3, 4],
            self.eval('[1, 2, 3, 4].delete(0, 0)'))

        self.assertEqual(
            [],
            self.eval('[1, 2, 3, 4].delete(0, -1)'))

    def test_insert(self):
        self.assertEqual(
            [1, 'a', 2],
            self.eval('[1, 2].insert(1, a)'))

        self.assertEqual(
            [1, ['a', 'b'], 2],
            self.eval('[1, 2].insert(1, [a, b])'))

        self.assertEqual(
            [1, 'a', 2],
            self.eval('[1, 2].insert(-1, a)'))

        self.assertEqual(
            [1, 2, 'a'],
            self.eval('[1, 2].insert(100, a)'))

        self.assertEqual(
            ['b', 'a'],
            self.eval('[].insert(0, a).insert(0, b)'))

        self.assertRaises(
            exceptions.NoMatchingMethodException,
            self.eval, 'set(a, b).insert(1, a)')

    def test_insert_iter(self):
        self.assertEqual(
            [1, 'a', 2],
            self.eval('[1, 2].select($).insert(1, a)'))

        self.assertEqual(
            [1, ['a', 'b'], 2],
            self.eval('[1, 2].select($).insert(1, [a, b])'))

        self.assertEqual(
            [1, 2],
            self.eval('[1, 2].select($).insert(-1, a)'))

        self.assertEqual(
            [1, 2, 'a'],
            self.eval('[1, 2].select($).insert(100, a)'))

        self.assertEqual(
            ['b', 'a'],
            self.eval('[].select($).insert(0, a).insert(0, b)'))

        self.assertEqual(
            ['a', 'a', 'b'],
            self.eval('set(a, b).orderBy($).insert(1, a)'))

    def test_insert_many(self):
        self.assertEqual(
            [1, 'a', 'b', 2],
            self.eval('[1, 2].insertMany(1, [a, b])'))

        self.assertEqual(
            ['a', 'b', 1, 2],
            self.eval('[1, 2].insertMany(-1, [a, b])'))

        self.assertEqual(
            [1, 2, 'a', 'b'],
            self.eval('[1, 2].insertMany(100, [a, b])'))

        self.assertEqual(
            ['a', 'a', 'b', 'b'],
            self.eval('[].insertMany(0, [a, b]).insertMany(1, [a, b])'))

    def test_replace(self):
        self.assertEqual(
            [None, 2, 3, 4],
            self.eval('[1, 2, 3, 4].replace(0, null)'))

        self.assertEqual(
            [None, 3, 4],
            self.eval('[1, 2, 3, 4].replace(0, null, 2)'))

        self.assertEqual(
            [1, 7],
            self.eval('[1, 2, 3, 4].replace(1, 7, -1)'))

        self.assertEqual(
            7,
            self.eval('set(1, 2, 3, 4).replace(1, 7)')[1])

        self.assertEqual(
            [1, 7, 3, 4],
            self.eval('set(4, 2, 3, 1).orderBy($).replace(1, 7)'))

    def test_replace_many(self):
        self.assertEqual(
            [7, 8, 2, 3, 4],
            self.eval('[1, 2, 3, 4].replaceMany(0, [7, 8])'))

        self.assertEqual(
            [7, 8, 3, 4],
            self.eval('[1, 2, 3, 4].replaceMany(0, [7, 8], 2)'))

        self.assertEqual(
            [1, 7, 8],
            self.eval('[1, 2, 3, 4].replaceMany(1, [7, 8], -1)'))

    def test_delete_dict(self):
        data = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        self.assertEqual(
            {'a': 1, 'd': 4},
            self.eval('$.delete(b, c)', data=data))

    def test_delete_all(self):
        data = {'a': 1, 'b': 2, 'c': 3, 'd': 4}
        self.assertEqual(
            {'a': 1, 'd': 4},
            self.eval('$.deleteAll([b, c])', data=data))

    def test_set(self):
        self.assertItemsEqual([2, 1, 3], self.eval('set(1, 2, 3, 2, 1)'))
        self.assertEqual([[1, 2, 3, 2, 1]], self.eval('set([1, 2, 3, 2, 1])'))
        self.assertEqual([], self.eval('set()'))
        self.assertEqual(
            [{'a': {'b': 'c'}}],
            self.eval('set({a => {b => c}})'))

    def test_set_from_iterator(self):
        self.assertItemsEqual([2, 1, 3], self.eval('set([1, 2, 3].select($))'))

    def test_to_set(self):
        self.assertItemsEqual(
            [2, 1, 3], self.eval('[1, 2, 3].select($).toSet()'))

        self.assertItemsEqual(
            [2, 1, 3], self.eval('[1, 2, 3].toSet()'))

    def test_set_len(self):
        self.assertEqual(3, self.eval('set(1, 2, 3).len()'))
        self.assertEqual(3, self.eval('len(set(1, 2, 3))'))

    def test_set_addition(self):
        self.assertItemsEqual(
            [4, 3, 2, 1],
            self.eval('set(1, 2, 3) + set(4, 2, 3)'))

        self.assertTrue(
            self.eval('isSet(set(1, 2, 3) + set(4, 2, 3))'))

    def test_set_union(self):
        self.assertItemsEqual(
            [4, 3, 2, 1],
            self.eval('set(1, 2, 3).union(set(4, 2, 3))'))

    def test_set_eq(self):
        self.assertTrue(self.eval('set(1, 2, 3) = set(3, 2, 1)'))
        self.assertFalse(self.eval('set(1, 2, 3) = set(1, 2, 3, 4)'))

    def test_set_neq(self):
        self.assertFalse(self.eval('set(1, 2, 3) != set(3, 2, 1)'))
        self.assertTrue(self.eval('set(1, 2, 3) != set(1, 2, 3, 4)'))

    def test_set_lt(self):
        self.assertTrue(self.eval('set(1, 2, 3) < set(1, 2, 3, 4)'))
        self.assertFalse(self.eval('set(1, 2, 3) < set(1, 2, 5)'))

    def test_set_gt(self):
        self.assertTrue(self.eval('set(1, 2, 3, 4) > set(1, 2, 3)'))
        self.assertFalse(self.eval('set(1, 2, 3) > set(1, 2, 3)'))

    def test_set_gte(self):
        self.assertFalse(self.eval('set(1, 2, 4) >= set(1, 2, 3)'))
        self.assertTrue(self.eval('set(1, 2, 3) >= set(1, 2, 3)'))

    def test_set_lte(self):
        self.assertFalse(self.eval('set(1, 2, 3) <= set(1, 2, 4)'))
        self.assertTrue(self.eval('set(1, 2, 3) <= set(1, 2, 3)'))

    def test_set_difference(self):
        self.assertItemsEqual(
            [4, 1],
            self.eval('set(1, 2, 3, 4).difference(set(2, 3))'))

    def test_set_subtraction(self):
        self.assertItemsEqual(
            [4, 1],
            self.eval('set(1, 2, 3, 4) - set(2, 3)'))

        self.assertTrue(
            self.eval('isSet(set(1, 2, 3, 4) - set(2, 3))'))

    def test_set_symmetric_difference(self):
        self.assertItemsEqual(
            [4, 1, 5],
            self.eval('set(1, 2, 3, 4).symmetricDifference(set(2, 3, 5))'))

    def test_set_add(self):
        self.assertItemsEqual(
            [4, 1, 2, 3],
            self.eval('set(1, 2, 3).add(4)'))

        self.assertItemsEqual(
            [4, 1, 2, 3, 5],
            self.eval('set(1, 2, 3).add(4, 5)'))

        self.assertItemsEqual(
            [1, 3, 2, [1, 2]],
            self.eval('set(1, 2, 3).add([1, 2])'))

        self.assertItemsEqual(
            [4, 1, None, 2, 3, 5],
            self.eval('set(1, 2, 3).add(4, 5, null)'))

        self.assertTrue(
            self.eval('isSet(set(1, 2, 3).add(4, 5, null))'))

    def test_set_remove(self):
        self.assertItemsEqual(
            [1, 3],
            self.eval('set(1, 2, 3).remove(2)'))

        self.assertItemsEqual(
            [3, None],
            self.eval('set(1, 2, null, 3).remove(1, 2, 5)'))

        self.assertItemsEqual(
            [3],
            self.eval('set(1, 2, null, 3).remove(1, 2, 5, null)'))

        self.assertItemsEqual(
            [1, 3, 2],
            self.eval('set(1, 2, 3, [1, 2]).remove([1, 2])'))

        self.assertTrue(
            self.eval('isSet(set(1, 2, 3, [1, 2]).remove([1, 2]))'))



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six


YAQLIZATION_ATTR = '__yaqlization__'


def yaqlize(class_or_object=None, yaqlize_attributes=True,
            yaqlize_methods=True, yaqlize_indexer=True,
            auto_yaqlize_result=False, whitelist=None, blacklist=None,
            attribute_remapping=None, blacklist_remapped_attributes=True):
    def func(something):
        if not hasattr(something, YAQLIZATION_ATTR):
            setattr(something, YAQLIZATION_ATTR, build_yaqlization_settings(
                yaqlize_attributes=yaqlize_attributes,
                yaqlize_methods=yaqlize_methods,
                yaqlize_indexer=yaqlize_indexer,
                auto_yaqlize_result=auto_yaqlize_result,
                whitelist=whitelist,
                blacklist=blacklist,
                attribute_remapping=attribute_remapping,
            ))
        return something
    if class_or_object is None:
        return func
    else:
        return func(class_or_object)


def get_yaqlization_settings(class_or_object):
    return getattr(class_or_object, YAQLIZATION_ATTR, None)


def is_yaqlized(class_or_object):
    return hasattr(class_or_object, YAQLIZATION_ATTR)


def build_yaqlization_settings(
        yaqlize_attributes=True, yaqlize_methods=True, yaqlize_indexer=True,
        auto_yaqlize_result=False, whitelist=None, blacklist=None,
        attribute_remapping=None, blacklist_remapped_attributes=True):
    whitelist = set(whitelist or [])
    blacklist = set(blacklist or [])
    attribute_remapping = attribute_remapping or {}
    if blacklist_remapped_attributes:
        for value in six.itervalues(attribute_remapping):
            if not isinstance(value, six.string_types):
                name = value[0]
            else:
                name = value
            blacklist.add(name)

    return {
        'yaqlizeAttributes': yaqlize_attributes,
        'yaqlizeMethods': yaqlize_methods,
        'yaqlizeIndexer': yaqlize_indexer,
        'autoYaqlizeResult': auto_yaqlize_result,
        'whitelist': whitelist,
        'blacklist': blacklist,
        'attributeRemapping': attribute_remapping
    }



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from yaql.language import exceptions
from yaql.language import expressions
from yaql.language import utils


class Parser(object):
    def __init__(self, lexer, yaql_operators, engine):
        self.tokens = lexer.tokens
        self._aliases = {}
        self._generate_operator_funcs(yaql_operators, engine)

    def _generate_operator_funcs(self, yaql_operators, engine):
        binary_doc = ''
        unary_doc = ''
        precedence_dict = {}

        for up, bp, op_name, op_alias in yaql_operators.operators.values():
            self._aliases[op_name] = op_alias
            if up:
                l = precedence_dict.setdefault(
                    (abs(up), 'l' if up > 0 else 'r'), [])
                l.append('UNARY_' + op_name if bp else op_name)
                unary_doc += ('value : ' if not unary_doc else '\n| ')
                spec_prefix = '{0} value' if up > 0 else 'value {0}'
                if bp:
                    unary_doc += (spec_prefix + ' %prec UNARY_{0}').format(
                        op_name)
                else:
                    unary_doc += spec_prefix.format(op_name)
            if bp:
                l = precedence_dict.setdefault(
                    (abs(bp), 'l' if bp > 0 else 'r'), [])
                if op_name == 'INDEXER':
                    l.extend(('LIST', 'INDEXER'))
                elif op_name == 'MAP':
                    l.append('MAP')
                else:
                    l.append(op_name)
                    binary_doc += ((
                        'value : ' if not binary_doc else '\n| ') +
                        'value {0} value'.format(op_name))

        # noinspection PyProtectedMember
        def p_binary(this, p):
            alias = this._aliases.get(p.slice[2].type)
            p[0] = expressions.BinaryOperator(p[2], p[1], p[3], alias)

        # noinspection PyProtectedMember
        def p_unary(this, p):
            if p[1] in yaql_operators.operators:
                alias = this._aliases.get(p.slice[1].type)
                p[0] = expressions.UnaryOperator(p[1], p[2], alias)
            else:
                alias = this._aliases.get(p.slice[2].type)
                p[0] = expressions.UnaryOperator(p[2], p[1], alias)

        p_binary.__doc__ = binary_doc
        self.p_binary = six.create_bound_method(p_binary, self)
        p_unary.__doc__ = unary_doc
        self.p_unary = six.create_bound_method(p_unary, self)

        precedence = []
        for i in range(1, len(precedence_dict) + 1):
            for oa in ('r', 'l'):
                value = precedence_dict.get((i, oa))
                if value:
                    precedence.append(
                        (('left',) if oa is 'l' else ('right',)) +
                        tuple(value)
                    )
        precedence.insert(0, ('left', ','))
        precedence.reverse()
        self.precedence = tuple(precedence)

        def p_value_call(this, p):
            """
            func : value '(' args ')'
            """
            arg = ()
            if len(p) > 4:
                arg = p[3]
            p[0] = expressions.Function('#call', p[1], *arg)

        if engine.allow_delegates:
            self.p_value_call = six.create_bound_method(p_value_call, self)

    @staticmethod
    def p_value_to_const(p):
        """
        value : QUOTED_STRING
              | NUMBER
              | TRUE
              | FALSE
              | NULL
        """
        p[0] = expressions.Constant(p[1])

    @staticmethod
    def p_keyword_constant(p):
        """
        value : KEYWORD_STRING
        """
        p[0] = expressions.KeywordConstant(p[1])

    @staticmethod
    def p_value_to_dollar(p):
        """
        value : DOLLAR
        """
        p[0] = expressions.GetContextValue(expressions.Constant(p[1]))

    @staticmethod
    def p_val_in_parenthesis(p):
        """
        value : '(' value ')'
        """
        p[0] = expressions.Wrap(p[2])

    @staticmethod
    def p_args(p):
        """
        args : arglist
             | named_arglist
             | arglist ',' named_arglist
             | incomplete_arglist ',' named_arglist
             |
        """
        if len(p) == 1:
            p[0] = []
        elif len(p) == 2:
            p[0] = p[1]
        else:
            p[0] = p[1] + p[3]

    @staticmethod
    def p_indexer(p):
        """
        value : value INDEXER args ']'
        """
        p[0] = expressions.IndexExpression(p[1], *p[3])

    @staticmethod
    def p_list(p):
        """
        value : INDEXER args ']' %prec LIST
        """
        p[0] = expressions.ListExpression(*p[2])

    @staticmethod
    def p_map(p):
        """
        value : MAP args '}'
        """
        p[0] = expressions.MapExpression(*p[2])

    @staticmethod
    def p_val_to_function(p):
        """
        value : func
        """
        p[0] = p[1]

    @staticmethod
    def p_named_arg_definition(p):
        """
        named_arg : value MAPPING value
        """
        p[0] = expressions.MappingRuleExpression(p[1], p[3])

    @staticmethod
    def p_arg_list(p):
        """
        arglist : value
                | ',' arglist
                | arglist ',' arglist
                | incomplete_arglist ',' arglist
        """
        if len(p) == 2:
            p[0] = [p[1]]
        elif len(p) == 3:
            p[0] = [utils.NO_VALUE] + p[2]
        elif len(p) == 4:
            p[0] = p[1] + p[3]

    @staticmethod
    def p_incomplete_arg_list(p):
        """
        incomplete_arglist : arglist ','
        """
        p[0] = p[1] + [utils.NO_VALUE]

    @staticmethod
    def p_named_arg_list(p):
        """
        named_arglist : named_arg
                      | named_arglist ',' named_arg
        """
        if len(p) == 2:
            p[0] = [p[1]]
        else:
            p[0] = p[1] + [p[3]]

    @staticmethod
    def p_function(p):
        """
        func : FUNC args ')'
        """
        arg = ()
        if len(p) > 3:
            arg = p[2]
        p[0] = expressions.Function(p[1], *arg)

    @staticmethod
    def p_error(p):
        if p:
            raise exceptions.YaqlGrammarException(
                p.lexer.lexdata, p.value, p.lexpos)
        else:
            raise exceptions.YaqlGrammarException(None, None, None)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
The module describes which operations can be done with datetime objects.
"""

import datetime
import time as python_time

from yaql.language import specs
from yaql.language import yaqltypes

from dateutil import parser
from dateutil import tz


DATETIME_TYPE = datetime.datetime
TIMESPAN_TYPE = datetime.timedelta
ZERO_TIMESPAN = datetime.timedelta()
UTCTZ = yaqltypes.DateTime.utctz


def _get_tz(offset):
    if offset is None:
        return None
    if offset == ZERO_TIMESPAN:
        return UTCTZ
    return tz.tzoffset(None, seconds(offset))


@specs.name('datetime')
@specs.parameter('year', int)
@specs.parameter('month', int)
@specs.parameter('day', int)
@specs.parameter('hour', int)
@specs.parameter('minute', int)
@specs.parameter('second', int)
@specs.parameter('microsecond', int)
@specs.parameter('offset', TIMESPAN_TYPE)
def build_datetime(year, month, day, hour=0, minute=0, second=0,
                   microsecond=0, offset=ZERO_TIMESPAN):
    """:yaql:datetime

    Returns datetime object built on year, month, day, hour, minute, second,
    microsecond, offset.

    :signature: datetime(year, month, day, hour => 0, minute => 0, second => 0,
                         microsecond => 0, offset => timespan(0))
    :arg year: number of years in datetime
    :argType year: integer between 1 and 9999 inclusive
    :arg month: number of months in datetime
    :argType month: integer between 1 and 12 inclusive
    :arg day: number of days in datetime
    :argType day: integer between 1 and number of days in given month
    :arg hour: number of hours in datetime, 0 by default
    :argType hour: integer between 0 and 23 inclusive
    :arg minute: number of minutes in datetime, 0 by default
    :argType minute: integer between 0 and 59 inclusive
    :arg second: number of seconds in datetime, 0 by default
    :argType second: integer between 0 and 59 inclusive
    :arg microsecond: number of microseconds in datetime, 0 by default
    :argType microsecond: integer between 0 and 1000000-1
    :arg offset: datetime offset in microsecond resolution, needed for tzinfo,
        timespan(0) by default
    :argType offset: timespan type
    :returnType: datetime object

    .. code::

        yaql> let(datetime(2015, 9, 29)) -> [$.year, $.month, $.day]
        [2015, 9, 29]
    """
    zone = _get_tz(offset)
    return DATETIME_TYPE(year, month, day, hour, minute, second,
                         microsecond, zone)


@specs.name('datetime')
@specs.parameter('timestamp', yaqltypes.Number())
@specs.parameter('offset', TIMESPAN_TYPE)
def datetime_from_timestamp(timestamp, offset=ZERO_TIMESPAN):
    """:yaql:datetime

    Returns datetime object built by timestamp.

    :signature: datetime(timestamp, offset => timespan(0))
    :arg timestamp: timespan object to represent datetime
    :argType timestamp: number
    :arg offset: datetime offset in microsecond resolution, needed for tzinfo,
        timespan(0) by default
    :argType offset: timespan type
    :returnType: datetime object

    .. code::

        yaql> let(datetime(1256953732)) -> [$.year, $.month, $.day]
        [2009, 10, 31]
    """
    zone = _get_tz(offset)
    return DATETIME_TYPE.fromtimestamp(timestamp, tz=zone)


@specs.name('datetime')
@specs.parameter('string', yaqltypes.String())
@specs.parameter('format__', yaqltypes.String(True))
def datetime_from_string(string, format__=None):
    """:yaql:datetime

    Returns datetime object built by string parsed with format.

    :signature: datetime(string, format => null)
    :arg string: string representing datetime
    :argType string: string
    :arg format: format for parsing input string which should be supported
        with C99 standard of format codes. null by default, which means
        parsing with Python dateutil.parser usage
    :argType format: string
    :returnType: datetime object

    .. code::

        yaql> let(datetime("29.8?2015")) -> [$.year, $.month, $.day]
        [2015, 8, 29]
        yaql> let(datetime("29.8?2015", "%d.%m?%Y"))->[$.year, $.month, $.day]
        [2015, 8, 29]
    """
    if not format__:
        result = parser.parse(string)
    else:
        result = DATETIME_TYPE.strptime(string, format__)
    if not result.tzinfo:
        return result.replace(tzinfo=UTCTZ)
    return result


@specs.name('timespan')
@specs.parameter('days', int)
@specs.parameter('hours', int)
@specs.parameter('minutes', int)
@specs.parameter('seconds', yaqltypes.Integer())
@specs.parameter('milliseconds', yaqltypes.Integer())
@specs.parameter('microseconds', yaqltypes.Integer())
def build_timespan(days=0, hours=0, minutes=0, seconds=0,
                   milliseconds=0, microseconds=0):
    """:yaql:timespan

    Returns timespan object with specified args.

    :signature: timespan(days => 0, hours => 0, minutes => 0, seconds => 0,
                         milliseconds => 0, microseconds => 0)
    :arg days: number of days in timespan, 0 by default
    :argType days: integer
    :arg hours: number of hours in timespan, 0 by default
    :argType hours: integer
    :arg minutes: number of minutes in timespan, 0 by default
    :argType minutes: integer
    :arg seconds: number of seconds in timespan, 0 by default
    :argType seconds: integer
    :arg milliseconds: number of microseconds in timespan, 0 by default
    :argType milliseconds: integer
    :arg microsecond: number of microseconds in timespan, 0 by default
    :argType microsecond: integer
    :returnType: timespan object

    .. code::

        yaql> timespan(days => 1, hours => 2, minutes => 3).hours
        26.05
    """
    return TIMESPAN_TYPE(
        days=days, hours=hours, minutes=minutes, seconds=seconds,
        milliseconds=milliseconds, microseconds=microseconds)


@specs.yaql_property(TIMESPAN_TYPE)
def microseconds(timespan):
    """:yaql:property microseconds

    Returns total microseconds in timespan.

    :signature: timespan.microseconds
    :returnType: integer

    .. code::

        yaql> timespan(seconds => 1).microseconds
        1000000
    """
    return (86400000000 * timespan.days +
            1000000 * timespan.seconds +
            timespan.microseconds)


@specs.yaql_property(TIMESPAN_TYPE)
def milliseconds(timespan):
    """:yaql:property milliseconds

    Returns total milliseconds in timespan.

    :signature: timespan.milliseconds
    :returnType: float

    .. code::

        yaql> timespan(seconds => 1).milliseconds
        1000.0
    """
    return microseconds(timespan) / 1000.0


@specs.yaql_property(TIMESPAN_TYPE)
def seconds(timespan):
    """:yaql:property seconds

    Returns total seconds in timespan.

    :signature: timespan.seconds
    :returnType: float

    .. code::

        yaql> timespan(minutes => 1).seconds
        60.0
    """
    return microseconds(timespan) / 1000000.0


@specs.yaql_property(TIMESPAN_TYPE)
def minutes(timespan):
    """:yaql:property minutes

    Returns total minutes in timespan.

    :signature: timespan.minutes
    :returnType: float

    .. code::

        yaql> timespan(hours => 2).minutes
        120.0
    """
    return microseconds(timespan) / 60000000.0


@specs.yaql_property(TIMESPAN_TYPE)
def hours(timespan):
    """:yaql:property hours

    Returns total hours in timespan.

    :signature: timespan.hours
    :returnType: float

    .. code::

        yaql> timespan(days => 2).hours
        48.0
    """
    return microseconds(timespan) / 3600000000.0


@specs.yaql_property(TIMESPAN_TYPE)
def days(timespan):
    """:yaql:property days

    Returns total days in timespan.

    :signature: timespan.days
    :returnType: float

    .. code::

        yaql> timespan(days => 2, hours => 48).days
        4.0
    """
    return microseconds(timespan) / 86400000000.0


def now(offset=ZERO_TIMESPAN):
    """:yaql:now

    Returns the current local date and time.

    :signature: now(offset => timespan(0))
    :arg offset: datetime offset in microsecond resolution, needed for tzinfo,
        timespan(0) by default
    :argType offset: timespan type
    :returnType: datetime

    .. code::

        yaql> let(now()) -> [$.year, $.month, $.day]
        [2016, 7, 18]
        yaql> now(offset=>localtz()).hour - now().hour
        3
    """
    zone = _get_tz(offset)
    return DATETIME_TYPE.now(tz=zone)


def localtz():
    """:yaql:localtz

    Returns local time zone in timespan object.

    :signature: localtz()
    :returnType: timespan object

    .. code::

        yaql> localtz().hours
        3.0
    """
    if python_time.daylight:
        return TIMESPAN_TYPE(seconds=-python_time.altzone)
    else:
        return TIMESPAN_TYPE(seconds=-python_time.timezone)


def utctz():
    """:yaql:utctz

    Returns UTC time zone in timespan object.

    :signature: utctz()
    :returnType: timespan object

    .. code::

        yaql> utctz().hours
        0.0
    """
    return ZERO_TIMESPAN


@specs.name('#operator_+')
@specs.parameter('dt', yaqltypes.DateTime())
@specs.parameter('ts', TIMESPAN_TYPE)
def datetime_plus_timespan(dt, ts):
    """:yaql:operator +

    Returns datetime object with added timespan.

    :signature: left + right
    :arg left: input datetime object
    :argType left: datetime object
    :arg right: input timespan object
    :argType right: timespan object
    :returnType: datetime object

    .. code::

        yaql> let(now() + timespan(days => 100)) -> $.month
        10
    """
    return dt + ts


@specs.name('#operator_+')
@specs.parameter('ts', TIMESPAN_TYPE)
@specs.parameter('dt', yaqltypes.DateTime())
def timespan_plus_datetime(ts, dt):
    """:yaql:operator +

    Returns datetime object with added timespan.

    :signature: left + right
    :arg left: input timespan object
    :argType left: timespan object
    :arg right: input datetime object
    :argType right: datetime object
    :returnType: datetime object

    .. code::

        yaql> let(timespan(days => 100) + now()) -> $.month
        10
    """
    return ts + dt


@specs.name('#operator_-')
@specs.parameter('dt', yaqltypes.DateTime())
@specs.parameter('ts', TIMESPAN_TYPE)
def datetime_minus_timespan(dt, ts):
    """:yaql:operator -

    Returns datetime object with subtracted timespan.

    :signature: left - right
    :arg left: input datetime object
    :argType left: datetime object
    :arg right: input timespan object
    :argType right: timespan object
    :returnType: datetime object

    .. code::

        yaql> let(now() - timespan(days => 100)) -> $.month
        4
    """
    return dt - ts


@specs.name('#operator_-')
@specs.parameter('dt1', yaqltypes.DateTime())
@specs.parameter('dt2', yaqltypes.DateTime())
def datetime_minus_datetime(dt1, dt2):
    """:yaql:operator -

    Returns datetime object dt1 with subtracted dt2.

    :signature: left - right
    :arg left: input datetime object
    :argType left: datetime object
    :arg right: datetime object to be subtracted
    :argType right: datetime object
    :returnType: timespan object

    .. code::

        yaql> let(now() - now()) -> $.microseconds
        -325
    """
    return dt1 - dt2


@specs.name('#operator_+')
@specs.parameter('ts1', TIMESPAN_TYPE)
@specs.parameter('ts2', TIMESPAN_TYPE)
def timespan_plus_timespan(ts1, ts2):
    """:yaql:operator +

    Returns sum of two timespan objects.

    :signature: left + right
    :arg left: input timespan object
    :argType left: timespan object
    :arg right: input timespan object
    :argType right: timespan object
    :returnType: timespan object

    .. code::

        yaql> let(timespan(days => 1) + timespan(hours => 12)) -> $.hours
        36.0
    """
    return ts1 + ts2


@specs.name('#operator_-')
@specs.parameter('ts1', TIMESPAN_TYPE)
@specs.parameter('ts2', TIMESPAN_TYPE)
def timespan_minus_timespan(ts1, ts2):
    """:yaql:operator -

    Returns timespan object with subtracted another timespan object.

    :signature: left - right
    :arg left: input timespan object
    :argType left: timespan object
    :arg right: input timespan object
    :argType right: timespan object
    :returnType: timespan object

    .. code::

        yaql> let(timespan(days => 1) - timespan(hours => 12)) -> $.hours
        12.0
    """
    return ts1 - ts2


@specs.name('#operator_>')
@specs.parameter('dt1', yaqltypes.DateTime())
@specs.parameter('dt2', yaqltypes.DateTime())
def datetime_gt_datetime(dt1, dt2):
    """:yaql:operator >

    Returns true if left datetime is strictly greater than right datetime,
    false otherwise.

    :signature: left > right
    :arg left: left datetime object
    :argType left: datetime object
    :arg right: right datetime object
    :argType right: datetime object
    :returnType: boolean

    .. code::

        yaql> datetime(2011, 11, 11) > datetime(2010, 10, 10)
        true
    """
    return dt1 > dt2


@specs.name('#operator_>=')
@specs.parameter('dt1', yaqltypes.DateTime())
@specs.parameter('dt2', yaqltypes.DateTime())
def datetime_gte_datetime(dt1, dt2):
    """:yaql:operator >=

    Returns true if left datetime is greater or equal to right datetime,
    false otherwise.

    :signature: left >= right
    :arg left: left datetime object
    :argType left: datetime object
    :arg right: right datetime object
    :argType right: datetime object
    :returnType: boolean

    .. code::

        yaql> datetime(2011, 11, 11) >= datetime(2011, 11, 11)
        true
    """
    return dt1 >= dt2


@specs.name('#operator_<')
@specs.parameter('dt1', yaqltypes.DateTime())
@specs.parameter('dt2', yaqltypes.DateTime())
def datetime_lt_datetime(dt1, dt2):
    """:yaql:operator <

    Returns true if left datetime is strictly less than right datetime,
    false otherwise.

    :signature: left < right
    :arg left: left datetime object
    :argType left: datetime object
    :arg right: right datetime object
    :argType right: datetime object
    :returnType: boolean

    .. code::

        yaql> datetime(2011, 11, 11) < datetime(2011, 11, 11)
        false
    """
    return dt1 < dt2


@specs.name('#operator_<=')
@specs.parameter('dt1', yaqltypes.DateTime())
@specs.parameter('dt2', yaqltypes.DateTime())
def datetime_lte_datetime(dt1, dt2):
    """:yaql:operator <=

    Returns true if left datetime is less or equal to right datetime,
    false otherwise.

    :signature: left <= right
    :arg left: left datetime object
    :argType left: datetime object
    :arg right: right datetime object
    :argType right: datetime object
    :returnType: boolean

    .. code::

        yaql> datetime(2011, 11, 11) <= datetime(2011, 11, 11)
        true
    """
    return dt1 <= dt2


@specs.name('#operator_>')
@specs.parameter('ts1', TIMESPAN_TYPE)
@specs.parameter('ts2', TIMESPAN_TYPE)
def timespan_gt_timespan(ts1, ts2):
    """:yaql:operator >

    Returns true if left timespan is strictly greater than right timespan,
    false otherwise.

    :signature: left > right
    :arg left: left timespan object
    :argType left: timespan object
    :arg right: right timespan object
    :argType right: timespan object
    :returnType: boolean

    .. code::

        yaql> timespan(hours => 2) > timespan(hours => 1)
        true
    """
    return ts1 > ts2


@specs.name('#operator_>=')
@specs.parameter('ts1', TIMESPAN_TYPE)
@specs.parameter('ts2', TIMESPAN_TYPE)
def timespan_gte_timespan(ts1, ts2):
    """:yaql:operator >=

    Returns true if left timespan is greater or equal to right timespan,
    false otherwise.

    :signature: left >= right
    :arg left: left timespan object
    :argType left: timespan object
    :arg right: right timespan object
    :argType right: timespan object
    :returnType: boolean

    .. code::

        yaql> timespan(hours => 24) >= timespan(days => 1)
        true
    """
    return ts1 >= ts2


@specs.name('#operator_<')
@specs.parameter('ts1', TIMESPAN_TYPE)
@specs.parameter('ts2', TIMESPAN_TYPE)
def timespan_lt_timespan(ts1, ts2):
    """:yaql:operator <

    Returns true if left timespan is strictly less than right timespan,
    false otherwise.

    :signature: left < right
    :arg left: left timespan object
    :argType left: timespan object
    :arg right: right timespan object
    :argType right: timespan object
    :returnType: boolean

    .. code::

        yaql> timespan(hours => 23) < timespan(days => 1)
        true
    """
    return ts1 < ts2


@specs.name('#operator_<=')
@specs.parameter('ts1', TIMESPAN_TYPE)
@specs.parameter('ts2', TIMESPAN_TYPE)
def timespan_lte_timespan(ts1, ts2):
    """:yaql:operator <=

    Returns true if left timespan is less or equal to right timespan,
    false otherwise.

    :signature: left <= right
    :arg left: left timespan object
    :argType left: timespan object
    :arg right: right timespan object
    :argType right: timespan object
    :returnType: boolean

    .. code::

        yaql> timespan(hours => 23) <= timespan(days => 1)
        true
    """
    return ts1 <= ts2


@specs.name('#operator_*')
@specs.parameter('ts', TIMESPAN_TYPE)
@specs.parameter('n', yaqltypes.Number())
def timespan_by_num(ts, n):
    """:yaql:operator *

    Returns timespan object built on timespan multiplied by number.

    :signature: left * right
    :arg left: timespan object
    :argType left: timespan object
    :arg right: number to multiply timespan
    :argType right: number
    :returnType: timespan

    .. code::

        yaql> let(timespan(hours => 24) * 2) -> $.hours
        48.0
    """
    return TIMESPAN_TYPE(microseconds=(microseconds(ts) * n))


@specs.name('#operator_*')
@specs.parameter('n', yaqltypes.Number())
@specs.parameter('ts', TIMESPAN_TYPE)
def num_by_timespan(n, ts):
    """:yaql:operator *

    Returns timespan object built on number multiplied by timespan.

    :signature: left * right
    :arg left: number to multiply timespan
    :argType left: number
    :arg right: timespan object
    :argType right: timespan object
    :returnType: timespan

    .. code::

        yaql> let(2 * timespan(hours => 24)) -> $.hours
        48.0
    """
    return TIMESPAN_TYPE(microseconds=(microseconds(ts) * n))


@specs.name('#operator_/')
@specs.parameter('ts1', TIMESPAN_TYPE)
@specs.parameter('ts2', TIMESPAN_TYPE)
def div_timespans(ts1, ts2):
    """:yaql:operator /

    Returns result of division of timespan microseconds by another timespan
    microseconds.

    :signature: left / right
    :arg left: left timespan object
    :argType left: timespan object
    :arg right: right timespan object
    :argType right: timespan object
    :returnType: float

    .. code::

        yaql> timespan(hours => 24) / timespan(hours => 12)
        2.0
    """
    return (0.0 + microseconds(ts1)) / microseconds(ts2)


@specs.name('#operator_/')
@specs.parameter('ts', TIMESPAN_TYPE)
@specs.parameter('n', yaqltypes.Number())
def div_timespan_by_num(ts, n):
    """:yaql:operator /

    Returns timespan object divided by number.

    :signature: left / right
    :arg left: left timespan object
    :argType left: timespan object
    :arg right: number to divide by
    :argType right: number
    :returnType: timespan object

    .. code::

        yaql> let(timespan(hours => 24) / 2) -> $.hours
        12.0
    """
    return TIMESPAN_TYPE(microseconds=(microseconds(ts) / n))


@specs.name('#unary_operator_-')
@specs.parameter('ts', TIMESPAN_TYPE)
def negative_timespan(ts):
    """:yaql:operator unary -

    Returns negative timespan.

    :signature: -arg
    :arg arg: input timespan object
    :argType arg: timespan object
    :returnType: timespan object

    .. code::

        yaql> let(-timespan(hours => 24)) -> $.hours
        -24.0
    """
    return -ts


@specs.name('#unary_operator_+')
@specs.parameter('ts', TIMESPAN_TYPE)
def positive_timespan(ts):
    """:yaql:operator unary +

    Returns timespan.

    :signature: +arg
    :arg arg: input timespan object
    :argType arg: timespan object
    :returnType: timespan object

    .. code::

        yaql> let(+timespan(hours => -24)) -> $.hours
        -24.0
    """
    return ts


@specs.yaql_property(DATETIME_TYPE)
def year(dt):
    """:yaql:property year

    Returns year of given datetime.

    :signature: datetime.year
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30).year
        2006
    """
    return dt.year


@specs.yaql_property(DATETIME_TYPE)
def month(dt):
    """:yaql:property month

    Returns month of given datetime.

    :signature: datetime.month
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30).month
        11
    """
    return dt.month


@specs.yaql_property(DATETIME_TYPE)
def day(dt):
    """:yaql:property day

    Returns day of given datetime.

    :signature: datetime.day
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30).day
        21
    """
    return dt.day


@specs.yaql_property(DATETIME_TYPE)
def hour(dt):
    """:yaql:property hour

    Returns hour of given datetime.

    :signature: datetime.hour
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30).hour
        16
    """
    return dt.hour


@specs.yaql_property(DATETIME_TYPE)
def minute(dt):
    """:yaql:property minute

    Returns minutes of given datetime.

    :signature: datetime.minute
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30).minute
        30
    """
    return dt.minute


@specs.yaql_property(DATETIME_TYPE)
def second(dt):
    """:yaql:property minute

    Returns seconds of given datetime.

    :signature: datetime.second
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30, 2).second
        2
    """
    return dt.second


@specs.yaql_property(DATETIME_TYPE)
def microsecond(dt):
    """:yaql:property microsecond

    Returns microseconds of given datetime.

    :signature: datetime.microsecond
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30, 2, 123).microsecond
        123
    """
    return dt.microsecond


@specs.yaql_property(yaqltypes.DateTime())
def date(dt):
    """:yaql:property date

    Returns datetime object with only year, month, day and tzinfo
    part of given datetime.

    :signature: datetime.date
    :returnType: datetime object

    .. code::

        yaql> let(datetime(2006, 11, 21, 16, 30, 2, 123).date) ->
            [$.year, $.month, $.day, $.hour]
        [2006, 11, 21, 0]
    """
    return DATETIME_TYPE(
        year=dt.year, month=dt.month, day=dt.day, tzinfo=dt.tzinfo)


@specs.yaql_property(yaqltypes.DateTime())
def time(dt):
    """:yaql:property time

    Returns timespan object built on datetime without year, month, day and
    tzinfo part of it.

    :signature: datetime.time
    :returnType: timespan object

    .. code::

        yaql> let(datetime(2006, 11, 21, 16, 30).time) -> [$.hours, $.minutes]
        [16.5, 990.0]
    """
    return dt - date(dt)


@specs.yaql_property(DATETIME_TYPE)
def weekday(dt):
    """:yaql:property weekday

    Returns the day of the week as an integer, Monday is 0 and Sunday is 6.

    :signature: datetime.weekday
    :returnType: integer

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30).weekday
        1
    """
    return dt.weekday()


@specs.yaql_property(yaqltypes.DateTime())
def utc(dt):
    """:yaql:property utc

    Returns datetime converted to UTC.

    :signature: datetime.utc
    :returnType: datetime object

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30, offset =>
            timespan(hours => 3)).utc.hour
        13
    """
    return dt - dt.utcoffset()


@specs.yaql_property(DATETIME_TYPE)
def offset(dt):
    """:yaql:property offset

    Returns offset of local time from UTC.

    :signature: datetime.offset
    :returnType: timespan

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30, offset =>
            timespan(hours => 3)).offset.hours
        3.0
    """
    return dt.utcoffset() or ZERO_TIMESPAN


@specs.yaql_property(DATETIME_TYPE)
def timestamp(dt):
    """:yaql:property timestamp

    Returns total seconds from datetime(1970, 1, 1) to
    datetime UTC.

    :signature: datetime.timestamp
    :returnType: float

    .. code::

        yaql> datetime(2006, 11, 21, 16, 30).timestamp
        1164126600.0
    """
    return (utc(dt) - DATETIME_TYPE(1970, 1, 1, tzinfo=UTCTZ)).total_seconds()


@specs.method
@specs.parameter('dt', yaqltypes.DateTime())
@specs.parameter('year', int)
@specs.parameter('month', int)
@specs.parameter('day', int)
@specs.parameter('hour', int)
@specs.parameter('minute', int)
@specs.parameter('second', int)
@specs.parameter('microsecond', int)
@specs.parameter('offset', TIMESPAN_TYPE)
def replace(dt, year=None, month=None, day=None, hour=None, minute=None,
            second=None, microsecond=None, offset=None):
    """:yaql:replace

    Returns datetime object with applied replacements.

    :signature: dt.replace(year => null, month => null, day => null,
                           hour => null, minute => null, second => null,
                           microsecond => null, offset => null)
    :receiverArg dt: input datetime object
    :argType dt: datetime object
    :arg year: number of years to replace, null by default which means
        no replacement
    :argType year: integer between 1 and 9999 inclusive
    :arg month: number of months to replace, null by default which means
        no replacement
    :argType month: integer between 1 and 12 inclusive
    :arg day: number of days to replace, null by default which means
        no replacement
    :argType day: integer between 1 and number of days in given month
    :arg hour: number of hours to replace, null by default which means
        no replacement
    :argType hour: integer between 0 and 23 inclusive
    :arg minute: number of minutes to replace, null by default which means
        no replacement
    :argType minute: integer between 0 and 59 inclusive
    :arg second: number of seconds to replace, null by default which means
        no replacement
    :argType second: integer between 0 and 59 inclusive
    :arg microsecond: number of microseconds to replace, null by default
        which means no replacement
    :argType microsecond: integer between 0 and 1000000-1
    :arg offset: datetime offset in microsecond resolution to replace, null
        by default which means no replacement
    :argType offset: timespan type
    :returnType: datetime object

    .. code::

        yaql> datetime(2015, 9, 29).replace(year => 2014).year
        2014
    """
    replacements = {}
    if year is not None:
        replacements['year'] = year
    if month is not None:
        replacements['month'] = month
    if day is not None:
        replacements['day'] = day
    if hour is not None:
        replacements['hour'] = hour
    if minute is not None:
        replacements['minute'] = minute
    if second is not None:
        replacements['second'] = second
    if microsecond is not None:
        replacements['microsecond'] = microsecond
    if offset is not None:
        replacements['tzinfo'] = _get_tz(offset)

    return dt.replace(**replacements)


@specs.method
@specs.parameter('dt', yaqltypes.DateTime())
@specs.parameter('format__', yaqltypes.String())
def format_(dt, format__):
    """:yaql:format

    Returns a string representing datetime, controlled by a format string.

    :signature: dt.format(format)
    :receiverArg dt: input datetime object
    :argType dt: datetime object
    :arg format: format string
    :argType format: string
    :returnType: string

    .. code::

        yaql> now().format("%A, %d. %B %Y %I:%M%p")
        "Tuesday, 19. July 2016 08:49AM"
    """
    return dt.strftime(format__)


def is_datetime(value):
    """:yaql:isDatetime

    Returns true if value is datetime object, false otherwise.

    :signature: isDatetime(value)
    :arg value: input value
    :argType value: any
    :returnType: boolean

    .. code::

        yaql> isDatetime(now())
        true
        yaql> isDatetime(datetime(2010, 10, 10))
        true
    """
    return isinstance(value, DATETIME_TYPE)


def is_timespan(value):
    """:yaql:isTimespan
    Returns true if value is timespan object, false otherwise.

    :signature: isTimespan(value)
    :arg value: input value
    :argType value: any
    :returnType: boolean

    .. code::

        yaql> isTimespan(now())
        false
        yaql> isTimespan(timespan())
        true
    """
    return isinstance(value, TIMESPAN_TYPE)


def register(context):
    functions = (
        build_datetime, build_timespan, datetime_from_timestamp,
        datetime_from_string, now, localtz, utctz, utc,
        days, hours, minutes, seconds, milliseconds, microseconds,
        datetime_plus_timespan, timespan_plus_datetime,
        datetime_minus_timespan, datetime_minus_datetime,
        timespan_plus_timespan, timespan_minus_timespan,
        datetime_gt_datetime, datetime_gte_datetime,
        datetime_lt_datetime, datetime_lte_datetime,
        timespan_gt_timespan, timespan_gte_timespan,
        timespan_lt_timespan, timespan_lte_timespan,
        negative_timespan, positive_timespan,
        timespan_by_num, num_by_timespan, div_timespans, div_timespan_by_num,
        year, month, day, hour, minute, second, microsecond, weekday,
        offset, timestamp, date, time, replace, format_, is_datetime,
        is_timespan
    )

    for func in functions:
        context.register_function(func)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql.tests


class TestRegex(yaql.tests.TestCase):
    def test_matches(self):
        self.assertTrue(self.eval("regex('a.b').matches(axb)"))
        self.assertFalse(self.eval("regex('a.b').matches(abx)"))

    def test_matches_string_method(self):
        self.assertTrue(self.eval("axb.matches('a.b')"))
        self.assertFalse(self.eval("abx.matches('a.b')"))

    def test_matches_operator_regex(self):
        self.assertTrue(self.eval("axb =~ regex('a.b')"))
        self.assertFalse(self.eval("abx =~ regex('a.b')"))

    def test_not_matches_operator_regex(self):
        self.assertFalse(self.eval("axb !~ regex('a.b')"))
        self.assertTrue(self.eval("abx !~ regex('a.b')"))

    def test_matches_operator_string(self):
        self.assertTrue(self.eval("axb =~ 'a.b'"))
        self.assertFalse(self.eval("abx =~ 'a.b'"))

    def test_not_matches_operator_string(self):
        self.assertFalse(self.eval("axb !~ 'a.b'"))
        self.assertTrue(self.eval("abx !~ 'a.b'"))

    def test_search(self):
        self.assertEqual(
            '24.16',
            self.eval(r"regex(`(\d+)\.?(\d+)?`).search('a24.16b')"))

    def test_search_with_selector(self):
        self.assertEqual(
            '24.16 = 24(2-4) + 16(5-7)',
            self.eval(
                r"regex(`(\d+)\.?(\d+)?`).search("r"'aa24.16bb', "
                r"format('{0} = {1}({2}-{3}) + {4}({5}-{6})', "
                r"$.value, $2.value, $2.start, $2.end, "
                r"$3.value, $3.start, $3.end))"))

    def test_search_all(self):
        self.assertEqual(
            ['24', '16'],
            self.eval(r"regex(`\d+`).searchAll('a24.16b')"))

    def test_search_all_with_selector(self):
        self.assertEqual(
            ['24!', '16!'],
            self.eval(r"regex(`\d+`).searchAll('a24.16b', $.value+'!')"))

    def test_split(self):
        self.assertEqual(
            ['Words', 'words', 'words', ''],
            self.eval(r"regex(`\W+`).split('Words, words, words.')"))
        self.assertEqual(
            ['Words', ', ', 'words', ', ', 'words', '.', ''],
            self.eval(r"regex(`(\W+)`).split('Words, words, words.')"))
        self.assertEqual(
            ['Words', 'words, words.'],
            self.eval(r"regex(`\W+`).split('Words, words, words.', 1)"))
        self.assertEqual(
            ['0', '3', '9'],
            self.eval(r"regex('[a-f]+', ignoreCase => true).split('0a3B9')"))

    def test_split_on_string(self):
        self.assertEqual(
            ['Words', 'words', 'words', ''],
            self.eval(r"'Words, words, words.'.split(regex(`\W+`))"))
        self.assertEqual(
            ['Words', ', ', 'words', ', ', 'words', '.', ''],
            self.eval(r"'Words, words, words.'.split(regex(`(\W+)`))"))
        self.assertEqual(
            ['Words', 'words, words.'],
            self.eval(r"'Words, words, words.'.split(regex(`\W+`), 1)"))
        self.assertEqual(
            ['0', '3', '9'],
            self.eval(r"'0a3B9'.split(regex('[a-f]+', ignoreCase => true))"))

    def test_replace(self):
        self.assertEqual(
            'axxbxx',
            self.eval(r"regex(`\d+`).replace(a12b23, xx)"))
        self.assertEqual(
            'axxb23',
            self.eval(r"regex(`\d+`).replace(a12b23, xx, 1)"))

    def test_replace_backref(self):
        self.assertEqual(
            'Foo_Bar_Foo',
            self.eval(r"regex(`([a-z0-9])([A-Z])`).replace("
                      "FooBarFoo, `\\1_\\2`)"))

    def test_replace_on_string(self):
        self.assertEqual(
            'axxbxx',
            self.eval(r"a12b23.replace(regex(`\d+`), xx)"))
        self.assertEqual(
            'axxb23',
            self.eval(r"a12b23.replace(regex(`\d+`), xx, 1)"))

    def test_replace_by(self):
        self.assertEqual(
            'axxbyy',
            self.eval(r"regex(`\d+`).replaceBy(a12b23, "
                      r"let(a => int($.value)) -> switch("
                      r"$a < 20 => xx, true => yy))"))

        self.assertEqual(
            'axxb23',
            self.eval(r"regex(`\d+`).replaceBy(a12b23, "
                      r"let(a => int($.value)) -> switch("
                      r"$a < 20 => xx, true => yy), 1)"))

    def test_replace_by_on_string(self):
        self.assertEqual(
            'axxbyy',
            self.eval(r"a12b23.replaceBy(regex(`\d+`), "
                      r"with(int($.value)) -> switch("
                      r"$ < 20 => xx, true => yy))"))

        self.assertEqual(
            'axxb23',
            self.eval(r"a12b23.replaceBy(regex(`\d+`), "
                      r"let(a => int($.value)) -> switch("
                      r"$a < 20 => xx, true => yy), 1)"))

    def test_escape_regex(self):
        self.assertEqual(
            '\\[',
            self.eval(r"escapeRegex('[')"))

    def test_is_regex(self):
        self.assertTrue(self.eval('isRegex(regex("a.b"))'))
        self.assertFalse(self.eval('isRegex(123)'))
        self.assertFalse(self.eval('isRegex(abc)'))



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
The Math module describes implemented math operations on numbers.
"""
import random

import six

from yaql.language import specs
from yaql.language import yaqltypes


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_+')
def binary_plus(left, right):
    """:yaql:operator +

    Returns the sum of left and right operands.

    :signature: left + right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType right: number
    :returnType: number

    .. code::

        yaql> 3 + 2
        5
    """
    return left + right


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_-')
def binary_minus(left, right):
    """:yaql:operator -

    Returns the difference between left and right.

    :signature: left - right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType right: number
    :returnType: number

    .. code::

        yaql> 3 - 2
        1
    """
    return left - right


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_*')
def multiplication(left, right):
    """:yaql:operator *

    Returns left multiplied by right.

    :signature: left * right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType right: number
    :returnType: number

    .. code::

        yaql> 3 * 2.5
        7.5
    """
    return left * right


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_/')
def division(left, right):
    """:yaql:operator /

    Returns left divided by right.

    :signature: left / right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType right: number
    :returnType: number

    .. code::

        yaql> 3 / 2
        1
        yaql> 3.0 / 2
        1.5
    """
    if isinstance(left, six.integer_types) and isinstance(
            right, six.integer_types):
        return left // right
    return left / right


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_mod')
def modulo(left, right):
    """:yaql:operator mod

    Returns left modulo right.

    :signature: left mod right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType right: number
    :returnType: number

    .. code::

        yaql> 3 mod 2
        1
    """
    return left % right


@specs.parameter('op', yaqltypes.Number())
@specs.name('#unary_operator_+')
def unary_plus(op):
    """:yaql:operator unary +

    Returns +op.

    :signature: +op
    :arg op: operand
    :argType op: number
    :returnType: number

    .. code::

        yaql> +2
        2
    """
    return +op


@specs.parameter('op', yaqltypes.Number())
@specs.name('#unary_operator_-')
def unary_minus(op):
    """:yaql:operator unary -

    Returns -op.

    :signature: -op
    :arg op: operand
    :argType op: number
    :returnType: number

    .. code::

        yaql> -2
        -2
    """
    return -op


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_>')
def gt(left, right):
    """:yaql:operator >

    Returns true if left is strictly greater than right, false otherwise.

    :signature: left > right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType left: number
    :returnType: boolean

    .. code::

        yaql> 3 > 2
        true
    """
    return left > right


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_>=')
def gte(left, right):
    """:yaql:operator >=

    Returns true if left is greater or equal to right, false otherwise.

    :signature: left >= right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType left: number
    :returnType: boolean

    .. code::

        yaql> 3 >= 3
        true
    """
    return left >= right


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_<')
def lt(left, right):
    """:yaql:operator <

    Returns true if left is strictly less than right, false otherwise.

    :signature: left < right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType left: number
    :returnType: boolean

    .. code::

        yaql> 3 < 2
        false
    """
    return left < right


@specs.parameter('left', yaqltypes.Number())
@specs.parameter('right', yaqltypes.Number())
@specs.name('#operator_<=')
def lte(left, right):
    """:yaql:operator <=

    Returns true if left is less or equal to right, false otherwise.

    :signature: left <= right
    :arg left: left operand
    :argType left: number
    :arg right: right operand
    :argType left: number
    :returnType: boolean

    .. code::

        yaql> 3 <= 3
        true
    """
    return left <= right


@specs.parameter('op', yaqltypes.Number())
def abs_(op):
    """:yaql:abs

    Returns the absolute value of a number.

    :signature: abs(op)
    :arg op: input value
    :argType op: number
    :returnType: number

    .. code::

        yaql> abs(-2)
        2
    """
    return abs(op)


def int_(value):
    """:yaql:int

    Returns an integer built from number, string or null value.

    :signature: int(value)
    :arg value: input value
    :argType value: number, string or null
    :returnType: integer

    .. code::

        yaql> int("2")
        2
        yaql> int(12.999)
        12
        yaql> int(null)
        0
    """
    if value is None:
        return 0
    return int(value)


def float_(value):
    """:yaql:float

    Returns a floating number built from number, string or null value.

    :signature: float(value)
    :arg value: input value
    :argType value: number, string or null
    :returnType: float

    .. code::

        yaql> float("2.2")
        2.2
        yaql> float(12)
        12.0
        yaql> float(null)
        0.0
    """
    if value is None:
        return 0.0
    return float(value)


def random_():
    """:yaql:random

    Returns the next random floating number from [0.0, 1.0).

    :signature: random()
    :returnType: float

    .. code::

        yaql> random()
        0.6039529924951869
    """
    return random.random()


def random__(from_, to_):
    """:yaql:random

    Returns the next random integer from [a, b].

    :signature: random(from, to)
    :arg from: left value for generating random number
    :argType from: integer
    :arg to: right value for generating random number
    :argType to: integer
    :returnType: integer

    .. code::

        yaql> random(1, 2)
        2
        yaql> random(1, 2)
        1
    """
    return random.randint(from_, to_)


@specs.parameter('left', int)
@specs.parameter('right', int)
def bitwise_and(left, right):
    """:yaql:bitwiseAnd

    Returns applied "bitwise and" to left and right integers.
    Each bit of the output is 1 if the corresponding bit of left AND right
    is 1, otherwise 0.

    :signature: bitwiseAnd(left, right)
    :arg left: left value
    :argType left: integer
    :arg right: right value
    :argType right: integer
    :returnType: integer

    .. code::

        yaql> bitwiseAnd(6, 12)
        4
    """
    return left & right


@specs.parameter('left', int)
@specs.parameter('right', int)
def bitwise_or(left, right):
    """:yaql:bitwiseOr

    Returns applied "bitwise or" to left and right numbers.
    Each bit of the output is 1 if the corresponding bit of left OR right
    is 1, otherwise 0.

    :signature: bitwiseOr(left, right)
    :arg left: left value
    :argType left: integer
    :arg right: right value
    :argType right: integer
    :returnType: integer

    .. code::

        yaql> bitwiseOr(6, 12)
        14
    """
    return left | right


@specs.parameter('left', int)
@specs.parameter('right', int)
def bitwise_xor(left, right):
    """:yaql:bitwiseXor

    Returns applied "bitwise exclusive or" to left and right numbers.
    Each bit of the output is equal to the sum of corresponding left and right
    bits mod 2.

    :signature: bitwiseXor(left, right)
    :arg left: left value
    :argType left: integer
    :arg right: right value
    :argType right: integer
    :returnType: integer

    .. code::

        yaql> bitwiseXor(6, 12)
        10
    """
    return left ^ right


@specs.parameter('arg', int)
def bitwise_not(arg):
    """:yaql:bitwiseNot

    Returns an integer where each bit is a reversed corresponding bit of arg.

    :signature: bitwiseNot(arg)
    :arg arg: input value
    :argType arg: integer
    :returnType: integer

    .. code::

        yaql> bitwiseNot(6)
        -7
    """
    return ~arg


@specs.parameter('value', int)
@specs.parameter('bits_number', int)
def shift_bits_right(value, bits_number):
    """:yaql:shiftBitsRight

    Shifts the bits of value right by the number of bits bitsNumber.

    :signature: shiftBitsRight(value, bitsNumber)
    :arg value: given value
    :argType value: integer
    :arg bitsNumber: number of bits
    :argType right: integer
    :returnType: integer

    .. code::

        yaql> shiftBitsRight(8, 2)
        2
    """
    return value >> bits_number


@specs.parameter('value', int)
@specs.parameter('bits_number', int)
def shift_bits_left(value, bits_number):
    """:yaql:shiftBitsLeft

    Shifts the bits of value left by the number of bits bitsNumber.

    :signature: shiftBitsLeft(value, bitsNumber)
    :arg value: given value
    :argType value: integer
    :arg bitsNumber: number of bits
    :argType right: integer
    :returnType: integer

    .. code::

        yaql> shiftBitsLeft(8, 2)
        32
    """
    return value << bits_number


@specs.parameter('a', nullable=True)
@specs.parameter('b', nullable=True)
@specs.inject('operator', yaqltypes.Delegate('#operator_>'))
def max_(a, b, operator):
    """:yaql:max

    Returns max from a and b.

    :signature: max(a, b)
    :arg a: input value
    :argType a: number
    :arg b: input value
    :argType b: number
    :returnType: number

    .. code::

        yaql> max(8, 2)
        8
    """
    if operator(b, a):
        return b
    return a


@specs.inject('operator', yaqltypes.Delegate('#operator_>'))
def min_(a, b, operator):
    """:yaql:min

    Returns min from a and b.

    :signature: min(a, b)
    :arg a: input value
    :argType a: number
    :arg b: input value
    :argType b: number
    :returnType: number

    .. code::

        yaql> min(8, 2)
        2
    """
    if operator(b, a):
        return a
    return b


@specs.parameter('a', yaqltypes.Number())
@specs.parameter('b', yaqltypes.Number())
@specs.parameter('c', yaqltypes.Number(nullable=True))
def pow_(a, b, c=None):
    """:yaql:pow

    Returns a to the power b modulo c.

    :signature: pow(a, b, c => null)
    :arg a: input value
    :argType a: number
    :arg b: power
    :argType b: number
    :arg c: modulo. null by default, which means no modulo is done after power.
    :argType c: integer
    :returnType: number

    .. code::

        yaql> pow(3, 2)
        9
        yaql> pow(3, 2, 5)
        4
    """
    return pow(a, b, c)


@specs.parameter('num', yaqltypes.Number())
def sign(num):
    """:yaql:sign

    Returns 1 if num > 0; 0 if num = 0; -1 if num < 0.

    :signature: sign(num)
    :arg num: input value
    :argType num: number
    :returnType: integer (-1, 0 or 1)

    .. code::

        yaql> sign(2)
        1
    """
    if num > 0:
        return 1
    elif num < 0:
        return -1
    return 0


@specs.parameter('number', yaqltypes.Number())
@specs.parameter('ndigits', int)
def round_(number, ndigits=0):
    """:yaql:round

    Returns a floating number rounded to ndigits after the decimal point.

    :signature: round(number, ndigits => 0)
    :arg number: input value
    :argType number: number
    :arg ndigits: with how many digits after decimal point to round.
        0 by default
    :argType ndigits: integer
    :returnType: number

    .. code::

        yaql> round(12.52)
        13
        yaql> round(12.52, 1)
        12.5
    """
    return round(number, ndigits)


def is_integer(value):
    """:yaql:isInteger

    Returns true if value is an integer number, otherwise false.

    :signature: isInteger(value)
    :arg value: input value
    :argType value: any
    :returnType: boolean

    .. code::

        yaql> isInteger(12.0)
        false
        yaql> isInteger(12)
        true
    """
    return isinstance(value, six.integer_types) and not isinstance(value, bool)


def is_number(value):
    """:yaql:isNumber

    Returns true if value is an integer or floating number, otherwise false.

    :signature: isNumber(value)
    :arg value: input value
    :argType value: any
    :returnType: boolean

    .. code::

        yaql> isNumber(12.0)
        true
        yaql> isNumber(12)
        true
    """
    return (isinstance(value, six.integer_types + (float,))
            and not isinstance(value, bool))


def register(context):
    context.register_function(binary_plus)
    context.register_function(binary_minus)
    context.register_function(multiplication)
    context.register_function(division)
    context.register_function(modulo)
    context.register_function(unary_plus)
    context.register_function(unary_minus)
    context.register_function(abs_)
    context.register_function(gt)
    context.register_function(gte)
    context.register_function(lt)
    context.register_function(lte)
    context.register_function(int_)
    context.register_function(float_)
    context.register_function(random_)
    context.register_function(random__)
    context.register_function(bitwise_and)
    context.register_function(bitwise_or)
    context.register_function(bitwise_not)
    context.register_function(bitwise_xor)
    context.register_function(shift_bits_left)
    context.register_function(shift_bits_right)
    context.register_function(max_)
    context.register_function(min_)
    context.register_function(pow_)
    context.register_function(sign)
    context.register_function(round_)
    context.register_function(is_integer)
    context.register_function(is_number)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import six

from yaql.language import exceptions
from yaql.language import expressions
from yaql.language import utils
from yaql.language import yaqltypes


def call(name, context, args, kwargs, engine, receiver=utils.NO_VALUE,
         data_context=None, use_convention=False, function_filter=None):

    if data_context is None:
        data_context = context

    if function_filter is None:
        function_filter = lambda fd, ctx: True

    if receiver is utils.NO_VALUE:
        predicate = lambda fd, ctx: fd.is_function and function_filter(fd, ctx)
    else:
        predicate = lambda fd, ctx: fd.is_method and function_filter(fd, ctx)

    all_overloads = context.collect_functions(
        name, predicate, use_convention=use_convention)

    if not all_overloads:
        if receiver is utils.NO_VALUE:
            raise exceptions.NoFunctionRegisteredException(name)
        else:
            raise exceptions.NoMethodRegisteredException(name, receiver)
    else:
        delegate = choose_overload(
            name, all_overloads, engine, receiver, data_context, args, kwargs)
        try:
            result = delegate()
            utils.limit_memory_usage(engine, (1, result))
            return result
        except StopIteration as e:
            six.reraise(
                exceptions.WrappedException,
                exceptions.WrappedException(e),
                sys.exc_info()[2])


def choose_overload(name, candidates, engine, receiver, context, args, kwargs):
    def raise_ambiguous():
        if receiver is utils.NO_VALUE:
            raise exceptions.AmbiguousFunctionException(name)
        else:
            raise exceptions.AmbiguousMethodException(name, receiver)

    def raise_not_found():
        if receiver is utils.NO_VALUE:
            raise exceptions.NoMatchingFunctionException(name)
        else:
            raise exceptions.NoMatchingMethodException(name, receiver)

    candidates2 = []
    lazy_params = None
    no_kwargs = None
    if receiver is not utils.NO_VALUE:
        args = (receiver,) + args
    for level in candidates:
        new_level = []
        for c in level:
            if no_kwargs is None:
                no_kwargs = c.no_kwargs
                args, kwargs = translate_args(no_kwargs, args, kwargs)
            elif no_kwargs != c.no_kwargs:
                raise_ambiguous()

            mapping = c.map_args(args, kwargs, context, engine)
            if mapping is None:
                continue
            pos, kwd = mapping
            lazy = set()
            for i, pos_arg in enumerate(pos):
                if isinstance(pos_arg.value_type, yaqltypes.LazyParameterType):
                    lazy.add(i)
            for key, value in six.iteritems(kwd):
                if isinstance(value.value_type, yaqltypes.LazyParameterType):
                    lazy.add(key)
            if lazy_params is None:
                lazy_params = lazy
            elif lazy_params != lazy:
                raise_ambiguous()
            new_level.append((c, mapping))
        if new_level:
            candidates2.append(new_level)

    if len(candidates2) == 0:
        raise_not_found()

    arg_evaluator = lambda i, arg: (
        arg(utils.NO_VALUE, context, engine)
        if (i not in lazy_params and isinstance(arg, expressions.Expression)
            and not isinstance(arg, expressions.Constant))
        else arg
    )

    args = tuple(arg_evaluator(i, arg) for i, arg in enumerate(args))
    for key, value in six.iteritems(kwargs):
        kwargs[key] = arg_evaluator(key, value)

    delegate = None
    winner_mapping = None
    for level in candidates2:
        for c, mapping in level:
            try:
                d = c.get_delegate(receiver, engine, context, args, kwargs)
            except exceptions.ArgumentException:
                pass
            else:
                if delegate is not None:
                    if _is_specialization_of(winner_mapping, mapping):
                        continue
                    elif not _is_specialization_of(mapping, winner_mapping):
                        raise_ambiguous()
                delegate = d
                winner_mapping = mapping
        if delegate is not None:
            break

    if delegate is None:
        raise_not_found()
    return lambda: delegate()


def translate_args(without_kwargs, args, kwargs):
    if without_kwargs:
        if len(kwargs) > 0:
            raise exceptions.ArgumentException(six.next(iter(kwargs)))
        return args, {}
    pos_args = []
    kw_args = {}
    for t in args:
        if isinstance(t, expressions.MappingRuleExpression):
            param_name = t.source
            if isinstance(param_name, expressions.KeywordConstant):
                param_name = param_name.value
            else:
                raise exceptions.MappingTranslationException()
            kw_args[param_name] = t.destination
        else:
            pos_args.append(t)
    for key, value in six.iteritems(kwargs):
        if key in kw_args:
            raise exceptions.MappingTranslationException()
        else:
            kw_args[key] = value
    return tuple(pos_args), kw_args


def _is_specialization_of(mapping1, mapping2):
    args_mapping1, kwargs_mapping1 = mapping1
    args_mapping2, kwargs_mapping2 = mapping2
    res = False

    for a1, a2 in six.moves.zip(args_mapping1, args_mapping2):
        if a2.value_type.is_specialization_of(a1.value_type):
            return False
        elif a1.value_type.is_specialization_of(a2.value_type):
            res = True

    for key, a1 in six.iteritems(kwargs_mapping1):
        a2 = kwargs_mapping2[key]
        if a2.value_type.is_specialization_of(a1.value_type):
            return False
        elif a1.value_type.is_specialization_of(a2.value_type):
            res = True

    return res



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from yaql.language import utils


class YaqlInterface(object):
    def __init__(self, context, engine, receiver=utils.NO_VALUE):
        self.__sender = receiver
        self.__engine = engine
        self.__context = context

    @property
    def context(self):
        return self.__context

    @property
    def engine(self):
        return self.__engine

    @property
    def sender(self):
        return self.__sender

    def on(self, receiver):
        return YaqlInterface(self.context, self.engine, receiver)

    def __getattr__(self, item):
        def stub(*args, **kwargs):
            context = self.context
            args = utils.convert_input_data(args)
            kwargs = utils.convert_input_data(kwargs)
            limit_func = context('#iter', self.engine)
            return utils.convert_output_data(
                context(item, self.engine, self.sender)(*args, **kwargs),
                limit_func, self.engine)
        return stub

    def __call__(self, __expression, *args, **kwargs):
        context = self.context.create_child_context()
        args = utils.convert_input_data(args)
        for i, arg_value in enumerate(args):
            context['$' + str(i + 1)] = arg_value

        kwargs = utils.convert_input_data(kwargs)
        for arg_name, arg_value in six.iteritems(kwargs):
            context['$' + arg_name] = arg_value

        parsed = self.engine(__expression)
        res = parsed.evaluate(context=context)
        limit_func = context('#iter', self.engine)
        return utils.convert_output_data(res, limit_func, self.engine)

    def __getitem__(self, item):
        return self.context[item]

    def __setitem__(self, key, value):
        self.context[key] = value



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import specs
from yaql.language import yaqltypes
import yaql.tests


class TestYaqlInterface(yaql.tests.TestCase):
    def test_call(self):
        def foo(yaql_interface):
            return yaql_interface('2+2')

        @specs.inject('yi', yaqltypes.YaqlInterface())
        def bar(yi):
            return yi('$a * $', 2, a=3)

        self.context.register_function(foo)
        self.context.register_function(bar)
        self.assertEqual(4, self.eval('foo()'))
        self.assertEqual(6, self.eval('bar()'))

    def test_function_call(self):
        def foo(yaql_interface):
            return yaql_interface.len([1, 2, 3])

        self.context.register_function(foo)
        self.assertEqual(3, self.eval('foo()'))

    def test_method_call(self):
        def foo(yaql_interface):
            return yaql_interface.on([1, 2, 3]).where(lambda i: i > 1)

        @specs.inject('yi', yaqltypes.YaqlInterface())
        def bar(yi):
            return yi.on([1, 2, 3]).select(yi.engine('$ * $'))

        self.context.register_function(foo)
        self.context.register_function(bar)
        self.assertEqual([2, 3], self.eval('foo()'))
        self.assertEqual([1, 4, 9], self.eval('bar()'))

    def test_data_access(self):
        def foo(yaql_interface):
            return yaql_interface[''], yaql_interface['key']

        self.context.register_function(foo)
        self.context['key'] = 'value'
        self.assertEqual(['test', 'value'], self.eval('foo()', data='test'))



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib
import operator
import pkgutil
import traceback
import types

import six
from docutils import nodes
from docutils.parsers import rst
from docutils import utils


TAG = ':yaql:'


def _get_modules_names(package):
    """Get names of modules in package"""

    return sorted(
        map(operator.itemgetter(1),
            pkgutil.walk_packages(package.__path__,
                                  '{0}.'.format(package.__name__))))


def _get_functions_names(module):
    """Get names of the functions in the current module"""

    return [name for name in dir(module) if
            isinstance(getattr(module, name, None), types.FunctionType)]


def write_method_doc(method, output):
    """Construct method documentation from a docstring.

    1) Strip TAG
    2) Embolden function name
    3) Add :callAs: after :signature:
    """

    msg = "Error: function {0} has no valid YAQL documentation."

    if method.__doc__:
        doc = method.__doc__
        try:
            # strip TAG
            doc = doc[doc.index(TAG) + len(TAG):]

            # embolden function name
            line_break = doc.index('\n')
            yaql_name = doc[:line_break]
            (emit_header, is_overload) = yield yaql_name
            if emit_header:
                output.write(yaql_name)
                output.write('\n')
                output.write('~' * len(yaql_name))
                output.write('\n')
            doc = doc[line_break:]

            # add :callAs: parameter
            try:
                signature_index = doc.index(':signature:')
                position = doc.index('    :', signature_index +
                                     len(':signature:'))
                if hasattr(method, '__yaql_function__'):
                    if (method.__yaql_function__.name and
                            'operator' in method.__yaql_function__.name):
                        call_as = 'operator'
                    elif (method.__yaql_function__.is_function and
                            method.__yaql_function__.is_method):
                        call_as = 'function or method'
                    elif method.__yaql_function__.is_method:
                        call_as = 'method'
                    else:
                        call_as = 'function'
                else:
                    call_as = 'function'

                call_as_str = '    :callAs: {0}\n'.format(call_as)
                text = doc[:position] + call_as_str + doc[position:]
            except ValueError:
                text = doc
            if is_overload:
                text = '*  ' + '\n   '.join(text.split('\n'))
                output.write(text)
            else:
                output.write(text)
        except ValueError:
            yield method.func_name
            output.write(msg.format(method.func_name))


def write_module_doc(module, output):
    """Generate and write rst document for module.

    Generate and write rst document for the single module.

    :parameter module: takes a Python module which should be documented.
    :type module: Python module

    :parameter output: takes file to which generated document will be written.
    :type output: file
    """
    functions_names = _get_functions_names(module)
    if module.__doc__:
        output.write(module.__doc__)
        output.write('\n')
    seq = []
    for name in functions_names:
        method = getattr(module, name)
        it = write_method_doc(method, output)
        try:
            name = six.next(it)
            seq.append((name, it))
        except StopIteration:
            pass
    seq.sort(key=operator.itemgetter(0))
    prev_name = None
    for i, item in enumerate(seq):
        name = item[0]
        emit_header = name != prev_name
        prev_name = name
        if emit_header:
            overload = i < len(seq) - 1 and seq[i + 1][0] == name
        else:
            overload = True

        try:
            item[1].send((emit_header, overload))
        except StopIteration:
            pass
        output.write('\n\n')
    output.write('\n')


def write_package_doc(package, output):
    """Writes rst document for the package.

    Generate and write rst document for the modules in the given package.

    :parameter package: takes a Python package which should be documented
    :type package: Python module

    :parameter output: takes file to which generated document will be written.
    :type output: file
    """

    modules = _get_modules_names(package)
    for module_name in modules:
        module = importlib.import_module(module_name)
        write_module_doc(module, output)


def generate_doc(source):
    try:
        package = importlib.import_module(source)
    except ImportError:
        return 'Error: No such module {0}'.format(source)
    out = six.StringIO()
    try:
        if hasattr(package, '__path__'):
            write_package_doc(package, out)
        else:
            write_module_doc(package, out)
        res = out.getvalue()
        return res

    except Exception as e:
        return '.. code-block:: python\n\n    Error: {0}\n    {1}\n\n'.format(
            str(e), '\n    '.join([''] + traceback.format_exc().split('\n')))


class YaqlDocNode(nodes.General, nodes.Element):
    source = None

    def __init__(self, source):
        self.source = source
        super(YaqlDocNode, self).__init__()


class YaqlDocDirective(rst.Directive):
    has_content = False
    required_arguments = 1

    def run(self):
        return [YaqlDocNode(self.arguments[0])]


def render(app, doctree, fromdocname):
    for node in doctree.traverse(YaqlDocNode):
        new_doc = utils.new_document('YAQL', doctree.settings)
        content = generate_doc(node.source)
        rst.Parser().parse(content, new_doc)
        node.replace_self(new_doc.children)


def setup(app):
    app.info('Loading the yaql documenter extension')
    app.add_node(YaqlDocNode)
    app.add_directive('yaqldoc', YaqlDocDirective)
    app.connect('doctree-resolved', render)
    return {'version': '0.1'}



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os.path
import pkg_resources

from yaql.language import contexts
from yaql.language import conventions
from yaql.language import factory
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes
from yaql.standard_library import boolean as std_boolean
from yaql.standard_library import branching as std_branching
from yaql.standard_library import collections as std_collections
from yaql.standard_library import common as std_common
from yaql.standard_library import date_time as std_datetime
from yaql.standard_library import math as std_math
from yaql.standard_library import queries as std_queries
from yaql.standard_library import regex as std_regex
from yaql.standard_library import strings as std_strings
from yaql.standard_library import system as std_system
from yaql.standard_library import yaqlized as std_yaqlized

_cached_expressions = {}
_cached_engine = None
_default_context = None


def detect_version():
    try:
        dist = pkg_resources.get_distribution('yaql')
        location = os.path.normcase(dist.location)
        this_location = os.path.normcase(__file__)
        if not this_location.startswith(os.path.join(location, 'yaql')):
            raise pkg_resources.DistributionNotFound()
        return dist.version
    except pkg_resources.DistributionNotFound:
        return 'Undefined (package was not installed with setuptools)'

__version__ = detect_version()


def _setup_context(data, context, finalizer, convention):
    if context is None:
        context = contexts.Context(
            convention=convention or conventions.CamelCaseConvention())

    if finalizer is None:
        @specs.parameter('iterator', yaqltypes.Iterable())
        @specs.name('#iter')
        def limit(iterator):
            return iterator

        @specs.inject('limiter', yaqltypes.Delegate('#iter'))
        @specs.inject('engine', yaqltypes.Engine())
        @specs.name('#finalize')
        def finalize(obj, limiter, engine):
            if engine.options.get('yaql.convertOutputData', True):
                return utils.convert_output_data(obj, limiter, engine)
            return obj

        context.register_function(limit)
        context.register_function(finalize)
    else:
        context.register_function(finalizer)

    if data is not utils.NO_VALUE:
        context['$'] = utils.convert_input_data(data)
    return context


def create_context(data=utils.NO_VALUE, context=None, system=True,
                   common=True, boolean=True, strings=True,
                   math=True, collections=True, queries=True,
                   regex=True, branching=True,
                   no_sets=False, finalizer=None, delegates=False,
                   convention=None, datetime=True, yaqlized=True):

    context = _setup_context(data, context, finalizer, convention)
    if system:
        std_system.register_fallbacks(context)
        context = context.create_child_context()
        std_system.register(context, delegates)
    if common:
        std_common.register(context)
    if boolean:
        std_boolean.register(context)
    if strings:
        std_strings.register(context)
    if math:
        std_math.register(context)
    if collections:
        std_collections.register(context, no_sets)
    if queries:
        std_queries.register(context)
    if regex:
        std_regex.register(context)
    if branching:
        std_branching.register(context)
    if datetime:
        std_datetime.register(context)
    if yaqlized:
        context = std_yaqlized.register(context)

    return context

YaqlFactory = factory.YaqlFactory


def eval(expression, data=None):
    global _cached_engine, _cached_expressions, _default_context

    if _cached_engine is None:
        _cached_engine = YaqlFactory().create()

    parsed_expression = _cached_expressions.get(expression)
    if parsed_expression is None:
        parsed_expression = _cached_engine(expression)
        _cached_expressions[expression] = parsed_expression

    if _default_context is None:
        _default_context = create_context()

    return parsed_expression.evaluate(
        data=data, context=_default_context.create_child_context())



# Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

import six

import yaql
from yaql.language import exceptions
from yaql.language import factory
from yaql.language import specs
from yaql.language import yaqltypes
from yaql import tests


class TestEngine(tests.TestCase):
    def test_parser_grammar(self):
        # replace stderr with cString to write to
        copy = sys.stderr
        sys.stderr = six.StringIO()
        try:
            debug_opts = dict(self.engine_options)
            debug_opts['yaql.debug'] = True
            yaql.factory.YaqlFactory().create(options=debug_opts)
            sys.stderr.seek(0)
            err_out = sys.stderr.read()
            self.assertEqual('Generating LALR tables\n', err_out)
        finally:
            # put stderr back
            sys.stderr = copy

    def test_no_function_registered(self):
        self.assertRaises(
            exceptions.NoFunctionRegisteredException,
            self.eval, 'kjhfksjdhfk()')

    def test_no_method_registered(self):
        self.assertRaises(
            exceptions.NoMethodRegisteredException,
            self.eval, '[1,2].kjhfksjdhfk($)')

    def test_no_matching_function(self):
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'len(1, 2, 3)')

    def test_mapping_translation_exception(self):
        self.context.register_function(
            lambda *args, **kwargs: 1, name='f')
        self.assertRaises(
            exceptions.MappingTranslationException,
            self.eval, 'f(2+2 => 4)')

    def test_no_matching_method(self):
        self.assertRaises(
            exceptions.NoMatchingMethodException,
            self.eval, '[1, 2].select(1, 2, 3)')

    def test_duplicate_parameters(self):
        def raises():
            @specs.parameter('p')
            @specs.parameter('p')
            def f(p):
                return p

        self.assertRaises(
            exceptions.DuplicateParameterDecoratorException,
            raises)

    def test_invalid_parameter(self):
        def raises():
            @specs.parameter('x')
            def f(p):
                return p

        self.assertRaises(
            exceptions.NoParameterFoundException,
            raises)

    def test_lexical_error(self):
        self.assertRaises(
            exceptions.YaqlLexicalException,
            self.eval, '1 ? 2')

    def test_grammar_error(self):
        self.assertRaises(
            exceptions.YaqlGrammarException,
            self.eval, '1 2')

        self.assertRaises(
            exceptions.YaqlGrammarException,
            self.eval, '(2')

    def test_invalid_method(self):
        self.assertRaises(
            exceptions.InvalidMethodException,
            self.context.register_function, lambda: 1, name='f', method=True)

        @specs.parameter('x', yaqltypes.Lambda())
        def func(x):
            return x

        self.assertRaises(
            exceptions.InvalidMethodException,
            self.context.register_function, func, name='f2', method=True)

    def test_function_definition(self):
        def func(a, b, *args, **kwargs):
            return a, b, args, kwargs

        fd = specs.get_function_definition(func)

        self.assertEqual(
            (1, 2, (5, 7), {'kw1': 'x', 'kw2': None}),
            fd(self.engine, self.context)(
                1, 2, 5, 7, kw1='x', kw2=None))

        self.assertEqual(
            (1, 5, (), {}),
            fd(self.engine, self.context)(1, b=5))

    def test_eval(self):
        self.assertEqual(
            120,
            yaql.eval(
                'let(inp => $) -> [1, 2, 3].select($ + $inp).reduce($1 * $2)',
                data=3)
        )

    def test_skip_args(self):
        def func(a=11, b=22, c=33):
            return a, b, c

        self.context.register_function(func)
        self.assertEqual([11, 22, 1], self.eval('func(,,1)'))
        self.assertEqual([0, 22, 1], self.eval('func(0,,1)'))
        self.assertEqual([11, 22, 33], self.eval('func()'))
        self.assertEqual([11, 1, 4], self.eval('func(,1, c=>4)'))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'func(,1, b=>4)')
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'func(,1,, c=>4)')

    def test_no_trailing_commas(self):
        self.assertRaises(exceptions.YaqlGrammarException,
                          self.eval, 'func(1,,)')
        self.assertRaises(exceptions.YaqlGrammarException,
                          self.eval, 'func(,1,)')
        self.assertRaises(exceptions.YaqlGrammarException,
                          self.eval, 'func(,,)')
        self.assertRaises(exceptions.YaqlGrammarException,
                          self.eval, 'func(,)')

    def test_no_varargs_after_kwargs(self):
        self.assertRaises(exceptions.YaqlGrammarException,
                          self.eval, 'func(x=>y, t)')
        self.assertRaises(exceptions.YaqlGrammarException,
                          self.eval, 'func(x=>y, ,t)')
        self.assertRaises(exceptions.YaqlGrammarException,
                          self.eval, 'func(a, x=>y, ,t)')

    def test_super(self):
        @specs.parameter('string', yaqltypes.String())
        @specs.inject('base', yaqltypes.Super())
        def len2(string, base):
            return 2 * base(string)

        context = self.context.create_child_context()
        context.register_function(len2, name='len')
        self.assertEqual(6, self.eval('len(abc)', context=context))

    def test_delegate_factory(self):
        @specs.parameter('name', yaqltypes.String())
        @specs.inject('__df__', yaqltypes.Delegate())
        def call_func(__df__, name, *args, **kwargs):
            return __df__(name, *args, **kwargs)

        context = self.context.create_child_context()

        context.register_function(call_func)
        self.assertEqual(
            [1, 2],
            self.eval('callFunc(list, 1, 2)', context=context))

        self.assertEqual(
            6,
            self.eval("callFunc('#operator_*', 2, 3)", context=context))

    def test_keyword_constant_function(self):
        @specs.parameter('arg', yaqltypes.Keyword())
        def foo(arg):
            return arg

        context = self.context.create_child_context()
        context.register_function(foo)

        self.assertEqual('qw', self.eval('foo(qw)', context=context))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', context=context, data='asd')
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo("q w")', context=context)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo(null)', context=context)

    def test_constant_function(self):
        @specs.parameter('arg', yaqltypes.Constant(True))
        def foo(arg):
            return arg

        context = self.context.create_child_context()
        context.register_function(foo)

        self.assertEqual('qw', self.eval('foo(qw)', context=context))
        self.assertEqual(123, self.eval('foo(123)', context=context))
        self.assertFalse(self.eval('foo(false)', context=context))
        self.assertIsNone(self.eval('foo(null)', context=context))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', context=context, data='asd')

    def test_string_constant_function(self):
        @specs.parameter('arg', yaqltypes.StringConstant())
        def foo(arg):
            return arg

        context = self.context.create_child_context()
        context.register_function(foo)

        self.assertEqual('qw', self.eval('foo(qw)', context=context))
        self.assertEqual('q w', self.eval('foo("q w")', context=context))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', context=context, data='asd')
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo(123)', context=context)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo(null)', context=context)

    def test_numeric_constant_function(self):
        @specs.parameter('arg', yaqltypes.NumericConstant())
        def foo(arg):
            return arg

        context = self.context.create_child_context()
        context.register_function(foo)

        self.assertEqual(123, self.eval('foo(123)', context=context))
        self.assertEqual(12.4, self.eval('foo(12.4)', context=context))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', context=context, data=5)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo("123")', context=context)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo(null)', context=context)

    def test_boolean_constant_function(self):
        @specs.parameter('arg', yaqltypes.BooleanConstant())
        def foo(arg):
            return arg

        context = self.context.create_child_context()
        context.register_function(foo)

        self.assertTrue(self.eval('foo(true)', context=context))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', context=context, data=True)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo("true")', context=context)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo(null)', context=context)

    def test_high_precedence_operator_insertion(self):
        # Test that insertion of high precedence operator doesn't
        # make turn ($.a)[0] into $.(a[0])
        engine_factory = factory.YaqlFactory(allow_delegates=True)
        engine_factory.insert_operator(
            None, True, ':',
            factory.OperatorType.BINARY_LEFT_ASSOCIATIVE, True)
        engine = engine_factory.create()
        data = {'a': [1]}
        expr = engine('$.a[0]')
        self.assertEqual(1, expr.evaluate(context=self.context, data=data))



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql.tests


class TestBranching(yaql.tests.TestCase):
    def test_switch(self):
        expr = 'switch($ < 10 => 1, $ >= 10 and $ < 100 => 2, $ >= 100 => 3)'
        self.assertEqual(3, self.eval(expr, data=123))
        self.assertEqual(2, self.eval(expr, data=50))
        self.assertEqual(1, self.eval(expr, data=-123))

    def test_select_case(self):
        expr = 'selectCase($ < 10, $ >= 10 and $ < 100)'
        self.assertEqual(2, self.eval(expr, data=123))
        self.assertEqual(1, self.eval(expr, data=50))
        self.assertEqual(0, self.eval(expr, data=-123))

    def test_select_all_cases(self):
        expr = 'selectAllCases($ < 10, $ > 5)'
        self.assertEqual([0], self.eval(expr, data=1))
        self.assertEqual([0, 1], self.eval(expr, data=7))
        self.assertEqual([1], self.eval(expr, data=12))

    def test_examine(self):
        expr = 'examine($ < 10, $ > 5)'
        self.assertEqual([True, False], self.eval(expr, data=1))
        self.assertEqual([True, True], self.eval(expr, data=7))
        self.assertEqual([False, True], self.eval(expr, data=12))

    def test_switch_case(self):
        expr = "$.switchCase('a', 'b', 'c')"
        self.assertEqual('a', self.eval(expr, data=0))
        self.assertEqual('b', self.eval(expr, data=1))
        self.assertEqual('c', self.eval(expr, data=3))
        self.assertEqual('c', self.eval(expr, data=30))
        self.assertEqual('c', self.eval(expr, data=-30))

    def test_coalesce(self):
        self.assertEqual(2, self.eval('coalesce($, 2)', data=None))
        self.assertEqual(1, self.eval('coalesce($, 2)', data=1))
        self.assertEqual(2, self.eval('coalesce($, $, 2)', data=None))
        self.assertIsNone(self.eval('coalesce($)', data=None))



#    Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from yaql.language import exceptions
from yaql import tests
from yaql import yaqlization


class TestYaqlization(tests.TestCase):
    def _get_sample_class(self):
        class D(object):
            d_attr = 777

        class C(object):
            def __init__(self):
                self.attr = 123

            def m_foo(self, arg1, arg2):
                return arg1 - arg2

            def bar(self, string):
                return string.upper()

            def get_d(self):
                return D()

            @staticmethod
            def static(string):
                return string.upper()

            @classmethod
            def clsmethod(cls, string):
                return string.upper()

            @property
            def prop(self):
                return self.attr

            def __getitem__(self, item):
                return item

        return C

    def test_method_call_yaqlized_object(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj)

        self.assertEqual(3, self.eval('$.m_foo(5, 2)', obj))
        self.assertEqual(3, self.eval('$.m_foo(5, arg2 => 2)', obj))
        self.assertEqual(3, self.eval('$.m_foo(arg2 => 2, arg1 => 6-1)', obj))
        self.assertEqual('A', self.eval('$.bar(a)', obj))
        self.assertEqual('B', self.eval('$.static(b)', obj))
        self.assertEqual('C', self.eval('$.clsmethod(c)', obj))
        self.assertRaises(
            exceptions.NoFunctionRegisteredException,
            self.eval, 'm_foo($, 5, 2)', obj)
        self.assertEqual(3, self.eval('$?.m_foo(5, 2)', obj))
        self.assertIsNone(self.eval('$?.m_foo(5, 2)', None))

    def test_method_call_yaqlized_class(self):
        cls = self._get_sample_class()
        yaqlization.yaqlize(cls)
        obj = cls()
        self.assertEqual(3, self.eval('$.m_foo(5, 2)', obj))

    def test_method_call_not_yaqlized(self):
        obj = self._get_sample_class()()
        self.assertRaises(
            exceptions.NoMethodRegisteredException,
            self.eval, '$.m_foo(5, 2)', obj)

    def test_method_call_forbidden(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, yaqlize_methods=False)
        self.assertRaises(
            exceptions.NoMethodRegisteredException,
            self.eval, '$.m_foo(5, 2)', obj)

    def test_property_access(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj)

        self.assertEqual(123, self.eval('$.attr', obj))
        self.assertEqual(123, self.eval('$.prop', obj))
        self.assertEqual(123, self.eval('$?.prop', obj))
        self.assertIsNone(self.eval('$?.prop', None))
        self.assertRaises(AttributeError, self.eval, '$.invalid', obj)

    def test_indexation(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj)

        self.assertEqual('key', self.eval('$[key]', obj))

    def test_method_call_whitelist_string(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, whitelist=['m_foo'])

        self.assertEqual(3, self.eval('$.m_foo(5, 2)', obj))
        self.assertRaises(AttributeError, self.eval, '$.bar(a)', obj)

    def test_method_call_whitelist_regexp(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, whitelist=[re.compile('^m_*')])

        self.assertEqual(3, self.eval('$.m_foo(5, 2)', obj))
        self.assertRaises(AttributeError, self.eval, '$.bar(a)', obj)

    def test_method_call_whitelist_callable(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, whitelist=[lambda t: t.startswith('m_')])

        self.assertEqual(3, self.eval('$.m_foo(5, 2)', obj))
        self.assertRaises(AttributeError, self.eval, '$.bar(a)', obj)

    def test_method_call_blacklist_string(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, blacklist=['m_foo'])

        self.assertRaises(AttributeError, self.eval, '$.m_foo(5, 2)', obj)
        self.assertEqual('A', self.eval('$.bar(a)', obj))

    def test_method_call_blacklist_regexp(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, blacklist=[re.compile('^m_*')])

        self.assertRaises(AttributeError, self.eval, '$.m_foo(5, 2)', obj)
        self.assertEqual('A', self.eval('$.bar(a)', obj))

    def test_method_call_blacklist_callable(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, blacklist=[lambda t: t.startswith('m_')])

        self.assertRaises(AttributeError, self.eval, '$.m_foo(5, 2)', obj)
        self.assertEqual('A', self.eval('$.bar(a)', obj))

    def test_property_access_blacklist(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, blacklist=['prop'])

        self.assertEqual(123, self.eval('$.attr', obj))
        self.assertRaises(AttributeError, self.eval, '$.prop', obj)

    def test_indexation_blacklist(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, blacklist=[lambda t: t.startswith('_')])

        self.assertEqual('key', self.eval('$[key]', obj))
        self.assertRaises(KeyError, self.eval, '$[_key]', obj)

    def test_auto_yaqlization(self):
        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj)

        self.assertRaises(
            exceptions.NoFunctionRegisteredException,
            self.eval, '$.get_d().d_attr', obj)

        obj = self._get_sample_class()()
        yaqlization.yaqlize(obj, auto_yaqlize_result=True)

        self.assertEqual(777, self.eval('$.get_d().d_attr', obj))

    def test_yaqlify_decorator(self):
        @yaqlization.yaqlize
        class C(object):
            attr = 555

        self.assertEqual(555, self.eval('$.attr', C()))

    def test_yaqlify_decorator_with_parameters(self):
        @yaqlization.yaqlize(yaqlize_attributes=True)
        class C(object):
            attr = 555

        self.assertEqual(555, self.eval('$.attr', C()))
